# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Callable, Dict, Generator, Optional, Tuple

import torch
import torchvision
from torch.nn import Parameter
from torch.optim import AdamW, Optimizer
from torch.optim.lr_scheduler import StepLR, _LRScheduler
from torchmetrics import Accuracy
from torchvision.transforms import transforms

from renate.data.data_module import RenateDataModule
from renate.models import RenateModule


class MyMNISTMLP(RenateModule):
    def __init__(self, num_hidden: int) -> None:
        # Model hyperparameters need to registered via RenateModule's
        # constructor, see documentation. Otherwise, this is a standard torch model.
        super().__init__(constructor_arguments={"num_hidden": num_hidden})
        self._fc1 = torch.nn.Linear(28 * 28, num_hidden)
        self._fc2 = torch.nn.Linear(num_hidden, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._fc1(x)
        x = torch.nn.functional.relu(x)
        return self._fc2(x)


def model_fn(model_state_url: Optional[str] = None) -> RenateModule:
    if model_state_url is None:
        # If no model state is given, we create the model from scratch with initial model
        # hyperparameters.
        model = MyMNISTMLP(num_hidden=100)
    else:
        # If a model state is passed, we reload the model using PyTorch's load_state_dict.
        # In this case, model hyperparameters are restored from the saved state.
        state_dict = torch.load(model_state_url)
        model = MyMNISTMLP.from_state_dict(state_dict)
    return model


class MyMNISTDataModule(RenateDataModule):
    def __init__(self, data_path: str, val_size: float, seed: int = 42) -> None:
        super().__init__(data_path, val_size=val_size, seed=seed)

    def prepare_data(self) -> None:
        # This is only to download the data. We separate downloading from the remaining set-up to
        # streamline data loading when using multiple training jobs during HPO.
        torchvision.datasets.MNIST(self._data_path, download=True)

    def setup(self) -> None:
        # This sets up train/val/test datasets, assuming data has already been downloaded.
        train_data = torchvision.datasets.MNIST(
            self._data_path,
            train=True,
            transform=transforms.ToTensor(),
            target_transform=transforms.Lambda(lambda x: torch.tensor(x, dtype=torch.long)),
        )
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = torchvision.datasets.MNIST(
            self._data_path,
            train=False,
            transform=transforms.ToTensor(),
            target_transform=transforms.Lambda(lambda x: torch.tensor(x, dtype=torch.long)),
        )


def data_module_fn(data_path: str, seed: int) -> RenateDataModule:
    return MyMNISTDataModule(val_size=0.2, seed=seed)


def train_transform() -> Callable:
    return torchvision.transforms.Compose(
        [torchvision.transforms.RandomCrop((28, 28), padding=4), torch.nn.Flatten()]
    )


def test_transform() -> Callable:
    return torch.nn.Flatten()


def buffer_transform() -> Callable:
    return torchvision.transforms.Compose(
        [
            torchvision.transforms.RandomCrop((28, 28), padding=4),
            torchvision.transforms.RandomRotation(degrees=15),
            torch.nn.Flatten(),
        ]
    )


def metrics_fn() -> Dict:
    return {"accuracy": Accuracy(task="multiclass", num_classes=10)}


def loss_fn() -> torch.nn.Module:
    return torch.nn.CrossEntropyLoss(reduction="none")


def optimizer_fn() -> Callable[[Generator[Parameter]], Optimizer]:
    return partial(AdamW, lr=0.01, weight_decay=0.0)


def lr_scheduler_fn() -> Tuple[Callable[[Optimizer], _LRScheduler], str]:
    return partial(StepLR, step_size=10, gamma=0.1), "epoch"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch
from torchmetrics import Metric

_DESCRIPTION = """
    In classification tasks, sample-wise inconsistencies between models appear as "negative flips":
    A new model incorrectly predicts the output for a test sample that was correctly
    classified by the old (reference) model. The fraction of the total number of negative flips
    and the total number of examples in the test set is called negative flip rate (NFR).
    Another metric for measuring regression is negative flip impact (NFI).
    Negative Flip Impact (NFI) measures the fraction of the total number of negative flips
    and the total number of errors in the test set. The difference between NFR and NFI
    is how they are normalized.
"""

_CITATION = """
    @article{Yan2021PositiveCongruentTT,
        title={Positive-Congruent Training: Towards Regression-Free Model Updates},
        author={Sijie Yan and Yuanjun Xiong and Kaustav Kundu and Shuo Yang
        and Siqi Deng and Meng Wang and Wei Xia and Stefano Soatto},
        journal={2021 IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR)},
        year={2021},
        pages={14294-14303}
    }
    @inproceedings{Xie2021RegressionBugs,
        title={Regression Bugs Are In Your Model! Measuring, Reducing and Analyzing Regressions In
            NLP Model Updates},
        author={Yuqing Xie and Yi{-}An Lai and Yuanjun Xiong and Yi Zhang and Shuo Yang and Stefano
            Soatto},
        journal={Proceedings of the 59th Annual Meeting of the Association for Computational
            Linguistics and the 11th International Joint Conference on Natural Language Processing
            (Volume 1: Long Papers)},
        year={2021},
        pages={6589-6602}
        }
    @article{Cai2022MeasuringAR,
        title={Measuring and Reducing Model Update Regression in Structured Prediction for NLP},
        author={Deng Cai and Elman Mansimov and Yi-An Lai and Yixuan Su and Lei Shu and Yi Zhang},
        journal={ArXiv},
        year={2022},
        volume={abs/2202.02976}
    }
"""


class NegativeFlipRateMetric(Metric):
    """Compute Negative Flip Rate (NFR) between new and old models' predictions,

    NFR = len((pred_old == labels) and (pred_new != labels)) / len(test_set).
    """

    full_state_update: bool = False
    higher_is_better: bool = False
    is_differentiable: bool = False

    def __init__(self):
        super().__init__()
        self.add_state("neg_flip_num", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, new_pred: torch.Tensor, old_pred: torch.Tensor, labels: torch.Tensor):
        """Update the metric.

        Args:
            new_pred: a 1-D torch tensor contains new model's predicted labels.
            old_pred: a 1-D torch tensor contains old model's predicted labels.
            labels: a 1-D torch tensor contains ground truth labels.
        """
        assert (
            new_pred.shape == old_pred.shape == labels.shape
        ), "dim of new_pred, old_pred, and labels are not matched!"
        assert len(new_pred) > 0, "input tensor is empty!"
        neg_flip_samples = (old_pred == labels) & (new_pred != labels)
        self.neg_flip_num += torch.sum(neg_flip_samples)
        self.total += len(labels)

    def compute(self):
        return self.neg_flip_num.float() / self.total


class NegativeFlipImpactMetric(Metric):
    """Compute Negative Flip Impact (NFI) between new and old models' predictions,

    NFI = len((pred_old == labels) and (pred_new != labels)) / len(pred_new != labels).
    """

    full_state_update: bool = False
    higher_is_better: bool = False
    is_differentiable: bool = False

    def __init__(self):
        super().__init__()
        self.add_state("new_num_errors", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("neg_flip_num", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, new_pred: torch.Tensor, old_pred: torch.Tensor, labels: torch.Tensor):
        """Updates the metric.

        Args:
            new_pred: a 1-D torch tensor contains new model's predicted labels.
            old_pred: a 1-D torch tensor contains old model's predicted labels.
            labels: a 1-D torch tensor contains ground truth labels.
        """
        assert (
            new_pred.shape == old_pred.shape == labels.shape
        ), "dim of new_pred, old_pred, and labels are not matched!"
        assert len(new_pred) > 0, "input tensor is empty!"
        neg_flip_samples = (old_pred == labels) & (new_pred != labels)
        self.neg_flip_num += torch.sum(neg_flip_samples)
        self.new_num_errors += torch.sum(new_pred != labels)

    def compute(self):
        return self.neg_flip_num.float() / self.new_num_errors


class PositiveFlipRateMetric(Metric):
    """Compute Positive Flip Rate (PFR) between new and old models' predictions,

    PFR = len((pred_old != labels) and (pred_new == labels)) / len(test_set).
    """

    full_state_update: bool = False
    higher_is_better: bool = True
    is_differentiable: bool = False

    def __init__(self):
        super().__init__()
        self.add_state("pos_flip_num", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, new_pred: torch.Tensor, old_pred: torch.Tensor, labels: torch.Tensor):
        """Updates the metric.

        Args:
            new_pred: a 1-D torch tensor contains new model's predicted labels.
            old_pred: a 1-D torch tensor contains old model's predicted labels.
            labels: a 1-D torch tensor contains ground truth labels.
        """
        assert (
            new_pred.shape == old_pred.shape == labels.shape
        ), "dim of new_pred, old_pred, and labels are not matched!"
        assert len(new_pred) > 0, "input tensor is empty!"
        pos_flip_samples = (old_pred != labels) & (new_pred == labels)
        self.pos_flip_num += torch.sum(pos_flip_samples)
        self.total += len(labels)

    def compute(self):
        return self.pos_flip_num.float() / self.total


class PositiveFlipImpactMetric(Metric):
    """Compute Positive Flip Impact (PFI) between new and old models' predictions,

    PFI = len((pred_old != labels) and (pred_new == labels)) / len(pred_new == labels).
    """

    full_state_update: bool = False
    higher_is_better: bool = True
    is_differentiable: bool = False

    def __init__(self):
        super().__init__()
        self.add_state("new_num_correct", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("pos_flip_num", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, new_pred: torch.Tensor, old_pred: torch.Tensor, labels: torch.Tensor):
        """Updates the metric.

        Args:
            new_pred: a 1-D torch tensor contains new model's predicted labels.
            old_pred: a 1-D torch tensor contains old model's predicted labels.
            labels: a 1-D torch tensor contains ground truth labels.
        """
        assert (
            new_pred.shape == old_pred.shape == labels.shape
        ), "dim of new_pred, old_pred, and labels are not matched!"
        assert len(new_pred) > 0, "input tensor is empty!"
        pos_flip_samples = (old_pred != labels) & (new_pred == labels)
        self.pos_flip_num += torch.sum(pos_flip_samples)
        self.new_num_correct += torch.sum(new_pred == labels)

    def compute(self):
        return self.pos_flip_num.float() / self.new_num_correct



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import shutil
from pathlib import Path
from tarfile import TarFile
from typing import List, Optional, Tuple, Union
from urllib.parse import urlparse
from zipfile import ZipFile

import boto3
import pandas as pd
import requests
from botocore.exceptions import ClientError
from pytorch_lightning.utilities.rank_zero import rank_zero_only

logger = logging.getLogger(__name__)


def get_aws_region() -> str:
    """Returns the name of the AWS region used during the execution."""
    return boto3.Session().region_name or "us-west-2"


def get_bucket() -> str:
    """Returns the default S3 bucket."""
    aws_account = boto3.client("sts").get_caller_identity().get("Account")
    return f"sagemaker-{get_aws_region()}-{aws_account}"


def is_s3_uri(uri: str) -> bool:
    """Checks if the uri is an S3 uri."""
    return urlparse(uri).scheme == "s3"


def _parse_s3_url(s3_url: str) -> Tuple[str, str]:
    parsed_url = urlparse(s3_url)
    if parsed_url.scheme == "s3":
        return parsed_url.netloc, parsed_url.path[1:]
    raise ValueError(f"{s3_url} is not an S3 URL.")


def _move_locally(
    src: Union[str, Path],
    dst: Union[str, Path],
    ignore_extensions: List[str] = [".sagemaker-uploading", ".sagemaker-uploaded"],
    copy: bool = False,
) -> None:
    """Moves files in directory or file to directory. If the files exist they are overwritten.

    Args:
        src: Source directory or file.
        dst: Target directory or file.
        ignore_extensions: List of extensions to ignore.
        copy: If `True`, copy instead of move.
    """
    if os.path.isfile(src):
        os.makedirs(dst, exist_ok=True)
        dst_file = os.path.join(dst, os.path.basename(src))
        if os.path.exists(dst_file):
            os.remove(dst_file)
        if copy:
            shutil.copy(src, dst)
        else:
            shutil.move(src, dst)
    for src_dir, _, files in os.walk(src):
        dst_dir = src_dir.replace(src, dst, 1)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for f in files:
            if f.endswith(tuple(ignore_extensions)):
                continue
            src_f = os.path.join(src_dir, f)
            dst_f = os.path.join(dst_dir, f)
            if os.path.exists(dst_f):
                os.remove(dst_f)
            if copy:
                shutil.copy(src_f, dst_f)
            else:
                shutil.move(src_f, dst_f)


def _move_to_s3(
    src: Union[str, Path],
    dst: Union[str, Path],
    ignore_extensions: List[str] = [".sagemaker-uploading", ".sagemaker-uploaded"],
    copy: bool = False,
) -> None:
    """Moves files in directory or file to directory or s3.

    If the files exist they are overwritten. The files in the local directory are deleted.

    Args:
        src: Local file or directory to move.
        dst: Target directory or s3 uri.
        ignore_extensions: List of extensions to ignore.
        copy: If `True`, copy instead of move.
    """
    if os.path.isfile(src):
        dst_file = os.path.join(dst, os.path.basename(src))
        upload_file_to_s3(src, dst_file)
        if not copy:
            os.remove(src)
    else:
        upload_folder_to_s3(src, dst, ignore_extensions=ignore_extensions)
        if not copy:
            shutil.rmtree(src)


def _move_to_uri(
    src: Union[Path, str],
    dst: str,
    ignore_extensions: List[str] = [".sagemaker-uploading", ".sagemaker-uploaded"],
    copy: bool = False,
) -> None:
    """Moves files in directory or file to directory or s3.

    If the files exist they are overwritten. The files in the local directory are deleted.

    Args:
        src: Local file or directory to move.
        dst: Target directory or s3 uri.
        ignore_extensions: List of extensions to ignore.
        copy: If `True`, copy instead of move.
    """
    if is_s3_uri(dst):
        _move_to_s3(src, dst, ignore_extensions=ignore_extensions, copy=copy)
    elif src != dst:
        _move_locally(src, dst, ignore_extensions=ignore_extensions, copy=copy)
    else:
        logging.warning(f"Source and destination are the same: {src}")


def move_to_uri(
    src: Union[Path, str],
    dst: str,
    ignore_extensions: List[str] = [".sagemaker-uploading", ".sagemaker-uploaded"],
) -> None:
    """Moves files in directory or file to directory or s3.

    If the files exist they are overwritten. The files in the local directory are deleted.

    Args:
        src: Local file or directory to move.
        dst: Target directory or s3 uri.
        ignore_extensions: List of extensions to ignore.
    """
    _move_to_uri(src=src, dst=dst, ignore_extensions=ignore_extensions, copy=False)


def copy_to_uri(
    src: Union[Path, str],
    dst: str,
    ignore_extensions: List[str] = [".sagemaker-uploading", ".sagemaker-uploaded"],
) -> None:
    """Copies files in directory or file to directory or s3.

    If the files exist they are overwritten. The files in the local directory are preserved.

    Args:
        src: Local directory to copy.
        dst: Target directory or s3 uri.
        ignore_extensions: List of extensions to ignore.
    """
    _move_to_uri(src=src, dst=dst, ignore_extensions=ignore_extensions, copy=True)


def maybe_download_from_s3(url: str, local_dir: Union[Path, str]) -> str:
    """Tries to download a file from S3."""
    try:
        src_bucket, src_object_name = _parse_s3_url(url)
        url = str(local_dir)
        download_folder_from_s3(
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            dst_dir=url,
        )
    except ValueError:
        pass
    return url


def download_folder_from_s3(
    src_bucket: str, src_object_name: Union[Path, str], dst_dir: Union[Path, str]
) -> None:
    """Downloads folder from S3 to local disk."""
    src_object_name = str(Path(src_object_name))
    s3_resource = boto3.resource("s3")
    bucket = s3_resource.Bucket(src_bucket)
    for obj in bucket.objects.filter(Prefix=src_object_name):
        dst_file = os.path.join(dst_dir, obj.key[len(src_object_name) + 1 :])
        if obj.key[-1] == "/":
            continue
        download_file_from_s3(src_bucket, obj.key, dst_file)


def upload_folder_to_s3(
    local_dir: Union[Path, str],
    s3_url: Optional[Union[Path, str]] = None,
    dst_bucket: Optional[str] = None,
    prefix: Optional[str] = None,
    ignore_extensions: List[str] = [".sagemaker-uploading", ".sagemaker-uploaded"],
) -> None:
    """Uploads all files within a local folder to s3.

    Args:
        local_dir: Folder containing files to be uploaded.
        s3_url: Full path to s3 location.
        dst_bucket: s3 bucket.
        prefix: Prefix for all s3 object names.
        ignore_extensions: List of extensions to ignore.
    """
    assert (
        s3_url is not None or dst_bucket is not None and prefix is not None
    ), "Either pass s3_url or both dst_bucket and prefix."
    if s3_url is not None:
        dst_bucket, prefix = _parse_s3_url(s3_url)
    local_dir = str(Path(local_dir))
    for current_folder, folders, files in os.walk(local_dir):
        for file_name in files:
            if file_name.endswith(tuple(ignore_extensions)):
                continue
            file_path = os.path.join(current_folder, file_name)
            object_name = os.path.join(prefix, current_folder[len(local_dir) + 1 :], file_name)
            upload_file_to_s3(file_path, dst_bucket=dst_bucket, dst_object_name=object_name)


def download_file_from_s3(
    src_bucket: str, src_object_name: Union[Path, str], dst: Union[Path, str]
) -> None:
    """Downloads file from S3 to local disk

    Args:
        src_bucket: Source S3 bucket
        src_object_name: Source S3 object
        dst: local destination
    """
    if isinstance(dst, str):
        dst = Path(dst)
    dst.parent.mkdir(parents=True, exist_ok=True)
    s3_client = boto3.client("s3")
    logger.info(f"Download file from s3://{src_bucket}/{src_object_name} to {dst}")
    with open(dst, "wb") as f:
        s3_client.download_fileobj(src_bucket, str(src_object_name), f)


def upload_file_to_s3(
    src: Union[Path, str],
    s3_url: Optional[Union[Path, str]] = None,
    dst_bucket: Optional[str] = None,
    dst_object_name: Optional[Union[Path, str]] = None,
) -> bool:
    """Upload a file to an S3 bucket

    Args:
        src: File to upload.
        s3_url: Full path to s3 location.
        dst_bucket: Destination S3 bucket
        dst_object_name: Destination S3 object

    Return:
        True if file was uploaded, else False
    """
    assert (
        s3_url is not None or dst_bucket is not None and dst_object_name is not None
    ), "Either pass s3_url or both dst_bucket and dst_object_name."
    if s3_url is not None:
        dst_bucket, dst_object_name = _parse_s3_url(s3_url)
    s3_client = boto3.client("s3")
    logger.info(f"Upload file from {src} to s3://{dst_bucket}/{dst_object_name}")
    try:
        s3_client.upload_file(str(src), dst_bucket, str(dst_object_name))
    except ClientError as e:
        logging.error(e)
        return False
    return True


def delete_file_from_s3(bucket: str, object_name: str) -> None:
    """Delete file from the S3 bucket

    Args:
        bucket: bucket in which the object (file) is stored
        object_name: object to be deleted
    """
    s3_client = boto3.client("s3")
    s3_client.delete_object(Bucket=bucket, Key=str(object_name))


def extract_file(dataset_name: str, data_path: Union[str, Path], file_name: str) -> None:
    """Extract .zip or .tar depending on the flag files into folder named with dataset name."""
    if file_name.endswith(".zip"):
        Extractor = ZipFile
    elif file_name.endswith(".tar"):
        Extractor = TarFile
    else:
        raise ValueError("Unknown compressed format extension. Only Zip/Tar supported.")
    with Extractor(os.path.join(data_path, dataset_name, file_name)) as f:
        f.extractall(os.path.join(data_path, dataset_name))


def download_file(
    dataset_name: str,
    data_path: Union[str, Path],
    src_bucket: str,
    src_object_name: str,
    url: str,
    file_name: str,
) -> None:
    """A helper function to download data from URL or s3."""
    if src_bucket is None:
        if not os.path.exists(os.path.join(data_path, dataset_name)):
            os.makedirs(os.path.join(data_path, dataset_name))
        with requests.get(os.path.join(url, file_name), allow_redirects=True, stream=True) as r:
            r.raise_for_status()
            with open(os.path.join(data_path, dataset_name, file_name), "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
    else:
        download_file_from_s3(
            src_bucket,
            os.path.join(src_object_name, file_name),
            os.path.join(data_path, dataset_name, file_name),
        )


def download_and_unzip_file(
    dataset_name: str,
    data_path: Union[str, Path],
    src_bucket: str,
    src_object_name: str,
    url: str,
    file_name: str,
) -> None:
    """A helper function to download data .zips and uncompress them."""
    download_file(dataset_name, data_path, src_bucket, src_object_name, url, file_name)
    extract_file(dataset_name, data_path, file_name)


def save_pandas_df_to_csv(df: pd.DataFrame, file_path: Union[str, Path]) -> pd.DataFrame:
    """A helper function to save pandas dataframe to a .csv.

    It guarantees that the saved dataframes across Renate are consistent.
    """
    df.to_csv(file_path, index=False)
    return df


@rank_zero_only
def unlink_file_or_folder(path: Path) -> None:
    """Function to remove files and folders.

    Unlink works for files, rmdir for empty folders, but not for non-empty ones. Hence a
    recursive solution.
    """
    if path.exists():
        if path.is_file():
            path.unlink(missing_ok=True)
        else:
            shutil.rmtree(path)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import argparse
import json
import os
import subprocess
from pathlib import Path
from sys import platform

import pandas as pd
import pytest

from renate import defaults

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        f"--test-file",
        type=str,
        required=True,
        help="Test suite to run.",
    )
    args = parser.parse_args()
    test_suite = "quick"
    current_folder = Path(os.path.dirname(__file__))
    configs_folder = current_folder / "configs"
    test_file = configs_folder / "suites" / test_suite / args.test_file
    if not test_file.is_file():
        raise FileNotFoundError(f"Unknown test file '{test_file}'.")
    with open(test_file) as f:
        test_config = json.load(f)
    job_name = f"{test_config['job_name']}-{defaults.current_timestamp()}"
    process = subprocess.Popen(
        [
            "python",
            current_folder / "run_experiment.py",
            "--scenario-file",
            configs_folder / "scenarios" / test_config["scenario"],
            "--model-file",
            configs_folder / "models" / test_config["model"],
            "--updater-file",
            configs_folder / "updaters" / test_config["updater"],
            "--dataset-file",
            configs_folder / "datasets" / test_config["dataset"],
            "--backend",
            test_config["backend"],
            "--job-name",
            job_name,
            "--test-suite",
            test_suite,
        ]
    )
    process.wait()
    expected_accuracy = test_config[f"expected_accuracy_{platform}"]
    num_updates = len(test_config["expected_accuracy_darwin"][0])
    result_file = (
        Path("tmp")
        / "renate-integration-tests"
        / test_suite
        / job_name
        / "0"
        / "logs"
        / f"metrics_summary_update_{num_updates - 1}.csv"
    )
    if result_file.exists():
        df = pd.read_csv(result_file)
        accuracies = [float(acc) for acc in list(df.iloc[-1])[1:]]
    else:
        accuracies = []
    assert any([pytest.approx(acc) == accuracies for acc in expected_accuracy]), accuracies



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Dict, Optional

import torch
from torchmetrics import Accuracy
from torchvision import transforms

import renate.defaults as defaults
from renate.benchmark.datasets.vision_datasets import TorchVisionDataModule
from renate.benchmark.models import ResNet18CIFAR
from renate.benchmark.scenarios import ClassIncrementalScenario, Scenario
from renate.models import RenateModule


def model_fn(model_state_url: Optional[str] = None) -> RenateModule:
    """Returns a model instance."""
    if model_state_url is None:
        model = ResNet18CIFAR()
    else:
        state_dict = torch.load(model_state_url)
        model = ResNet18CIFAR.from_state_dict(state_dict)
    return model


def data_module_fn(data_path: str, chunk_id: int, seed: int = defaults.SEED) -> Scenario:
    """Returns a class-incremental scenario instance.

    The transformations passed to prepare the input data are required to convert the data to
    PyTorch tensors.
    """
    data_module = TorchVisionDataModule(
        data_path,
        dataset_name="CIFAR10",
        val_size=0.2,
        seed=seed,
    )
    class_incremental_scenario = ClassIncrementalScenario(
        data_module=data_module,
        groupings=((0, 1, 2, 3, 4), (5, 6, 7, 8, 9)),
        chunk_id=chunk_id,
    )
    return class_incremental_scenario


def train_transform() -> Callable:
    """Returns a transform function to be used in the training."""
    return transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2615)),
        ]
    )


def test_transform() -> Callable:
    """Returns a transform function to be used for validation or testing."""
    return transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2615))


def buffer_transform() -> Callable:
    """Returns a transform function to be used in the Memory Buffer."""
    return train_transform()


def loss_fn() -> torch.nn.Module:
    return torch.nn.CrossEntropyLoss(reduction="none")


def metrics_fn() -> Dict:
    return {"accuracy": Accuracy(task="multiclass", num_classes=10)}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import boto3
from syne_tune.backend.sagemaker_backend.sagemaker_utils import get_execution_role

from renate.training import run_training_job

config_space = {
    "optimizer": "SGD",
    "momentum": 0.9,
    "weight_decay": 0.0,
    "learning_rate": 0.001,
    "alpha": 0.5,
    "batch_size": 64,
    "batch_memory_frac": 0.5,
    "memory_size": 300,
    "loss_normalization": 0,
    "loss_weight": 0.5,
}

if __name__ == "__main__":
    AWS_ID = boto3.client("sts").get_caller_identity().get("Account")
    AWS_REGION = "us-west-2"  # use your AWS preferred region here

    run_training_job(
        config_space=config_space,
        mode="min",
        metric="val_loss",
        updater="ER",  # we train with Experience Replay
        max_epochs=5,
        config_file="renate_config.py",
        # For this example, we can train on two binary movie review datasets: "rotten_tomatoes" and
        # "imdb". Set chunk_id to [0, 1] to switch between the two.
        chunk_id=0,
        # replace the url below with a different one if you already ran it and you want to avoid
        # overwriting
        output_state_url=f"s3://sagemaker-{AWS_REGION}-{AWS_ID}/renate-training-nlp-finetuning/",
        # uncomment the line below only if you already created a model with this script and you want
        # to update it
        # input_state_url=f"s3://sagemaker-{AWS_REGION}-{AWS_ID}/renate-training-nlp-finetuning/",
        backend="sagemaker",  # run on SageMaker, select "local" to run this locally
        role=get_execution_role(),
        instance_count=1,
        instance_type="ml.g4dn.xlarge",
        job_name="renate-training-nlp-finetuning",
        devices=1,
        strategy="deepspeed_stage_2",
        precision="32",
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from renate.benchmark.experimentation import execute_experiment_job, experiment_config_file


config_space = {
    "updater": "DER",
    "optimizer": "SGD",
    "momentum": 0.0,
    "weight_decay": 0.0,
    "learning_rate": 0.03,
    "alpha": 0.2,
    "beta": 0.5,
    "batch_size": 64,
    "batch_memory_frac": 0.5,
    "memory_size": 500,
    "max_epochs": 50,
    "loss_normalization": 0,
    "loss_weight": 1.0,
    "model_name": "ResNet18CIFAR",
    "scenario_name": "ClassIncrementalScenario",
    "dataset_name": "CIFAR10",
    "val_size": 0,
    "groupings": ((0, 1), (2, 3), (4, 5), (6, 7), (8, 9)),
    "num_outputs": 10,
}

for seed in range(10):
    execute_experiment_job(
        backend="local",
        config_file=experiment_config_file(),
        config_space=config_space,
        experiment_outputs_url=f"results/{seed}/",
        mode="max",
        metric="val_accuracy",
        num_updates=5,
        seed=seed,
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch
from torch.utils.data import ConcatDataset, TensorDataset

from renate import defaults
from renate.updaters.experimental.repeated_distill import (
    RepeatedDistillationModelUpdater,
    double_distillation_loss,
)


def test_dmc_runs_end_to_end():
    data = []
    classes_available = [0, 1, 2]

    mlp = pytest.helpers.get_renate_module_mlp(
        num_inputs=10, num_outputs=len(classes_available), hidden_size=256, num_hidden_layers=2
    )

    for i in classes_available:
        ds = torch.utils.data.TensorDataset(
            torch.mul(torch.ones(200, 10), i),
            torch.randint(low=i, high=i + 1, size=(200,)),
        )

        data.append(ds)

    val = ConcatDataset(data)
    updater = RepeatedDistillationModelUpdater(
        loss_fn=pytest.helpers.get_loss_fn(),
        optimizer=pytest.helpers.get_partial_optimizer(),
        model=mlp,
        memory_size=300,
        batch_size=20,
        max_epochs=5,
        accelerator="cpu",
    )

    for i in range(len(data)):
        updater.update(data[i], val_dataset=val, task_id=defaults.TASK_ID)


@pytest.mark.parametrize("dataset_size", [50, 200])
@pytest.mark.parametrize("memory_size", [10, 100])
def test_dmc_memory_size_after_update(memory_size, dataset_size):
    model = pytest.helpers.get_renate_module_mlp(
        num_inputs=10, num_outputs=3, hidden_size=20, num_hidden_layers=1
    )
    model_updater = RepeatedDistillationModelUpdater(
        loss_fn=pytest.helpers.get_loss_fn(),
        optimizer=pytest.helpers.get_partial_optimizer(),
        model=model,
        memory_size=memory_size,
        max_epochs=1,
        accelerator="cpu",
    )
    datasets = [
        TensorDataset(
            torch.randn(dataset_size, 10), torch.randint(low=0, high=3, size=(dataset_size,))
        )
        for _ in range(3)
    ]
    assert len(model_updater._learner._memory_buffer) == 0
    for i, dataset in enumerate(datasets):
        model_updater.update(dataset, task_id=defaults.TASK_ID)
        assert len(model_updater._learner._memory_buffer) == min(
            memory_size, (i + 1) * dataset_size
        )


@pytest.mark.parametrize("provide_folder", [True, False])
def test_dmc_model_updater(tmpdir, provide_folder):
    model, train_dataset, test_data = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    model_updater = RepeatedDistillationModelUpdater(
        model,
        loss_fn=pytest.helpers.get_loss_fn(),
        optimizer=pytest.helpers.get_partial_optimizer(),
        memory_size=50,
        max_epochs=1,
        output_state_folder=defaults.output_state_folder(tmpdir) if provide_folder else None,
        accelerator="cpu",
    )
    y_hat_before_train = model(test_data, task_id=defaults.TASK_ID)
    model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    y_hat_after_train = model(test_data, task_id=defaults.TASK_ID)
    assert y_hat_before_train.shape[0] == y_hat_after_train.shape[0]
    assert not torch.allclose(y_hat_before_train, y_hat_after_train)


def test_continuation_of_training_with_dmc_model_updater(tmpdir):
    model, train_dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    state_url = defaults.input_state_folder(tmpdir)
    model_updater = RepeatedDistillationModelUpdater(
        model,
        loss_fn=pytest.helpers.get_loss_fn(),
        optimizer=pytest.helpers.get_partial_optimizer(),
        memory_size=50,
        max_epochs=1,
        output_state_folder=state_url,
        accelerator="cpu",
    )
    model = model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    model_updater = RepeatedDistillationModelUpdater(
        model,
        loss_fn=pytest.helpers.get_loss_fn(),
        optimizer=pytest.helpers.get_partial_optimizer(),
        memory_size=50,
        max_epochs=1,
        input_state_folder=state_url,
        accelerator="cpu",
    )
    model_updater.update(train_dataset, task_id=defaults.TASK_ID)


def test_dmc_loss():
    a = torch.ones(10, 3)
    b = torch.zeros(10, 3)

    loss = double_distillation_loss(a, b)
    expected_loss = torch.mul(
        torch.ones(
            10,
        ),
        0.5,
    )

    assert torch.allclose(loss, expected_loss)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
from typing import Callable, Dict, List, Optional, Union

import torch
import torchmetrics
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.loggers.logger import Logger
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.data.datasets import _TransformedDataset
from renate.models import RenateModule
from renate.utils.distributed_strategies import create_strategy
from renate.utils.misc import int_or_str


class Evaluator(LightningModule, abc.ABC):
    """A general Evaluator module for collection of quantitative metrics on the test dataset.

    This is an abstract interface which can be called with respect to a PyTorch Lightning `Trainer`.
    and its `.test()` function. It collects quantitative observations with respect to a single
    dataset. The metrics that are being collected are defined in the `create_metrics` function.

    Args:
        model: A `RenateModule` to be evaluated.
        batch_size: The batch size to be used when creating the test data loader.
        transform: The transformation applied for evaluation.
        target_transform: The target transformation applied for evaluation.
        logged_metrics: Metrics logged additional to the default ones.
    """

    def __init__(
        self,
        model: RenateModule,
        batch_size: int,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
    ) -> None:
        super().__init__()
        self._model = model
        self._model.deregister_hooks()
        self._batch_size = batch_size
        self._transform = transform
        self._target_transform = target_transform
        self._metric_collection = torchmetrics.MetricCollection(logged_metrics)

    def on_model_test_start(
        self,
        test_dataset: Dataset,
        test_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> DataLoader:
        """Called before a model test starts."""
        test_dataset = _TransformedDataset(
            test_dataset,
            transform=self._transform,
            target_transform=self._target_transform,
        )
        self._task_id = task_id
        return DataLoader(
            test_dataset,
            batch_size=self._batch_size,
            shuffle=False,
            pin_memory=True,
            collate_fn=test_collate_fn,
        )

    def test_step(self, batch: List[torch.Tensor], batch_idx: int) -> None:
        """PyTorch Lightning function to perform the test step."""
        x, y = batch
        outputs = self(x)
        self._metric_collection(outputs, y)

    @abc.abstractmethod
    def forward(self, x, task_id: Optional[str] = None) -> torch.Tensor:
        """Forward pass of the model.

        Task ID can be used to specify, for example, the output head to perform the evaluation with
        a specific data Chunk ID. Here, the `task_id` is used only to compute the test metrics.
        """
        pass

    def on_test_epoch_end(self) -> None:
        """PyTorch Lightning function to perform at the end of test loop.

        Logs the metrics and resets the metric collection.
        """
        self.log_dict(self._metric_collection.compute(), on_step=False, on_epoch=True)
        self._metric_collection.reset()


class ClassificationEvaluator(Evaluator):
    """A classification Evaluator module for collection of quantitative metrics on the test
    dataset.
    """

    def forward(self, x, task_id: Optional[str] = None) -> torch.Tensor:
        """Forward pass of the model.

        Task ID can be used to specify, for example, the output head to perform the evaluation with
        a specific data Chunk ID. Here, the `task_id` is used only to compute the test metrics.
        """
        if task_id is None:
            task_id = self._task_id
        return self._model.get_logits(x, task_id=task_id)


def evaluate(
    model: RenateModule,
    test_dataset: Union[List[Dataset], Dataset],
    test_collate_fn: Optional[Callable] = None,
    task_id: Union[List[str], str] = defaults.TASK_ID,
    batch_size: int = defaults.BATCH_SIZE,
    transform: Optional[Callable] = None,
    target_transform: Optional[Callable] = None,
    logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
    logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
    accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
    devices: Optional[int] = None,
    strategy: str = defaults.DISTRIBUTED_STRATEGY,
    precision: str = defaults.PRECISION,
) -> Dict[str, List[float]]:
    """Evaluate the model on the test dataset or a set of test datasets corresponding to distinct
    tasks.

    If the `test_dataset` are specified as a list of datasets, it is assumed to be ordered.
    Similarly, in a case the `task_id` are specified as a list, it is assumed to be ordered. A task
    ID list can be used to set specific model part to be used, for example, an output head with some
    specific test dataset in the input sequence.

    Args:
        model: A `RenateModule` to be evaluated.
        test_dataset: The test dataset(s) to be evaluated.
        test_collate_fn: collate_fn used in the DataLoader.
        task_id: The task id(s) of the test dataset(s).
        batch_size: The batch size to be used when creating the test data loader.
        transform: The transformation applied for evaluation.
        target_transform: The target transformation applied for evaluation.
        logged_metrics: Metrics logged additional to the default ones.
        logger: Logger used by PyTorch Lightning to log intermediate results.
        accelerator: Accelerator used by PyTorch Lightning to train the model.
        devices: Devices used by PyTorch Lightning to train the model. If the devices flag is not
            defined, it will assume devices to be "auto" and fetch the `auto_device_count` from the
            `accelerator`.
        strategy: Name of the distributed training strategy to use.
            `More details <https://lightning.ai/docs/pytorch/stable/extensions/strategy.html>`__
        precision: Type of bit precision to use.
            `More details <https://lightning.ai/docs/pytorch/stable/common/precision_basic.html>`__
    """
    if isinstance(test_dataset, Dataset):
        test_dataset = [test_dataset]

    if isinstance(task_id, str):
        task_id = [task_id] * len(test_dataset)

    assert len(task_id) == len(test_dataset)

    evaluator = ClassificationEvaluator(
        model=model,
        batch_size=batch_size,
        transform=transform,
        target_transform=target_transform,
        logged_metrics=logged_metrics,
    )

    trainer = Trainer(
        accelerator=accelerator,
        devices=devices,
        logger=logger,
        enable_checkpointing=False,
        enable_progress_bar=False,
        strategy=create_strategy(devices, strategy),
        precision=int_or_str(precision),
    )

    results = {}
    for i in range(len(test_dataset)):
        test_loader = evaluator.on_model_test_start(test_dataset[i], test_collate_fn, task_id[i])
        trainer.test(
            evaluator,
            test_loader,
        )
        for metric_name, value in trainer.logged_metrics.items():
            if metric_name not in results:
                results[metric_name] = []
            results[metric_name].append(value.item())
    return results



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import math
from typing import Any, Iterator, List, Optional, Set, Tuple, Union

import torch
from torch.utils.data import BatchSampler, Dataset, Sampler, SubsetRandomSampler, random_split
from transformers import BatchEncoding

from renate import defaults
from renate.types import NestedTensors


def reinitialize_model_parameters(model: torch.nn.Module) -> None:
    """Reinitializes the parameters of a model.

    This relies on all submodules of `model` implementing a method `reset_parameters()`. This
    is implemented for core `torch.nn` layers, but this may not be the case for custom
    implementations of exotic layers. A warning is logged for modules that do not implement
    `reset_parameters()`.

    The actual logic of reinitializing parameters depends on the type of layer. It may affect the
    module's buffers (non-trainable parameters, e.g., batch norm stats) as well.

    Args:
        model: The model to be reinitialized.
    """
    for module in model.modules():
        # Skip modules without any parameters of their own.
        if not list(module.parameters(recurse=False)) and not list(module.buffers(recurse=False)):
            continue
        try:
            module.reset_parameters()
        except AttributeError:
            logging.warning(f"Skipping module {module} while resetting parameters.")


def get_generator(seed: Optional[int] = None) -> torch.Generator:
    """Provides a torch.Generator for the given seed.

    torch.default_generator is returned if seed is None.
    """
    if seed is None:
        return torch.default_generator
    rng = torch.Generator(device="cpu")
    rng.manual_seed(seed)
    return rng


def randomly_split_data(
    dataset: Dataset, proportions: List[float], seed: int = defaults.SEED
) -> List[Dataset]:
    """Randomly splits a dataset into chunks."""
    rng = get_generator(seed)
    split_sizes = _proportions_into_sizes(proportions, len(dataset))
    return random_split(dataset, split_sizes, generator=rng)


def _proportions_into_sizes(proportions: List[float], size: int) -> List[int]:
    """A helper function to convert chunk proportions into sizes.

    In case of rounding doubts, any remaining samples are appended to the last split size.
    """
    assert math.isclose(sum(proportions), 1.0), sum(proportions)
    sizes = [round(proportion * size) for proportion in proportions[:-1]]
    sizes.append(size - sum(sizes))
    return sizes


def move_tensors_to_device(tensors: NestedTensors, device: torch.device) -> NestedTensors:
    """Moves a collection of tensors to `device`.

    The collection `tensors` can be a nested structure of tensors, tuples, lists, and dicts.
    """
    if isinstance(tensors, (BatchEncoding, torch.Tensor)):
        return tensors.to(device)
    elif isinstance(tensors, tuple):
        return tuple(move_tensors_to_device(t, device) for t in tensors)
    # We need to include lists here as well, since collate_fn sometimes turns tuples into lists.
    # See https://github.com/pytorch/pytorch/issues/48419.
    elif isinstance(tensors, list):
        return [move_tensors_to_device(t, device) for t in tensors]
    elif isinstance(tensors, dict):
        return {key: move_tensors_to_device(t, device) for key, t in tensors.items()}
    else:
        raise TypeError(
            "Expected `tensors` to be a nested structure of tensors, tuples, list and dict; "
            f"discovered {type(tensors)}."
        )


def get_length_nested_tensors(batch: NestedTensors) -> torch.Size:
    """Given a NestedTensor, return its length.

    Assumes that the first axis in each element is the same.
    """
    if isinstance(batch, torch.Tensor):
        return batch.shape[0]
    if isinstance(batch, tuple):
        return batch[0].shape[0]
    if isinstance(batch, dict):
        return batch[next(iter(batch.keys()))].shape[0]


def cat_nested_tensors(
    nested_tensors: Union[Tuple[NestedTensors], List[NestedTensors]], axis: int = 0
) -> NestedTensors:
    """Concatenates the two NestedTensors.

    Equivalent of PyTorch's ``cat`` function for ``NestedTensors``.

    Args:
        nested_tensors: Tensors to be concatenated.
        axis: Concatenation axis.
    """
    if isinstance(nested_tensors[0], torch.Tensor):
        return torch.cat(nested_tensors, axis)
    if isinstance(nested_tensors[0], tuple):
        return tuple(
            cat_nested_tensors(nested_tensor, axis) for nested_tensor in zip(*nested_tensors)
        )
    if isinstance(nested_tensors[0], dict):
        return {
            key: cat_nested_tensors([nested_tensor[key] for nested_tensor in nested_tensors], axis)
            for key in nested_tensors[0]
        }


def unique_classes(dataset: torch.utils.data.Dataset) -> Set[int]:
    """Compute the unique class ids in a dataset.

    Args:
        dataset: Instance of Torch dataset.
    """
    from renate.memory.buffer import DataBuffer  # to avoid circular import

    if isinstance(dataset, DataBuffer):
        label_element = lambda elem: elem[0][1]
    else:
        label_element = lambda elem: elem[1]

    unique_values = set()
    for ind in range(len(dataset)):
        label = label_element(dataset[ind])
        unique_values.add(label.item())

    return unique_values


def complementary_indices(num_outputs: int, valid_classes: Set[int]) -> List[int]:
    """Compute the asymmetric difference between the two arguments

    Args:
        num_outputs: An integer of total number of classes the model can output.
        valid_classes: A set of integers of valid classes.
    """
    return [class_idx for class_idx in range(num_outputs) if class_idx not in valid_classes]


class ConcatRandomSampler(Sampler[List[int]]):
    """Sampler for sampling batches from ConcatDatasets.

    Each sampled batch is composed of batches of different BatchSamplers with the specified
    batch sizes and ranges.

    To clarify the behavior, we provide a little example.
    ``dataset_lengths = [5, 2]``
    ``batch_sizes = [3, 1]``

    With this setting, we have a set of indices A={0..4} and B={5,6} for the two datasets.
    The total batch size will be exactly 4. The first three elements are in that batch are
    elements of A, the last an element of B.
    An example batch could be ``[3, 1, 0, 6]``.

    Since we always provide a batch size of exactly ` sum(batch_sizes)``, we drop the last
    batch.


    Args:
        dataset_lengths: The length for the different datasets.
        batch_sizes: Batch sizes used for specific datasets.
        complete_dataset_iteration: Provide an index to indicate over which dataset to fully
            iterate. By default, stops whenever iteration is complete for any dataset.
        generator: Generator used in sampling.
        sampler: Lightning automatically passes a DistributedSamplerWrapper. Only used as an
            indicator that we are in the distributed case.
    """

    def __init__(
        self,
        dataset_lengths: List[int],
        batch_sizes: List[int],
        complete_dataset_iteration: Optional[int] = None,
        generator: Any = None,
        sampler: Sampler = None,
    ) -> None:
        self.batch_sizes = batch_sizes
        self.complete_dataset_iteration = complete_dataset_iteration
        self.subset_samplers = []
        data_start_idx = 0
        num_batches = []
        rank = 0 if sampler is None else sampler.rank
        num_replicas = 1 if sampler is None else sampler.num_replicas
        for dataset_length, batch_size in zip(dataset_lengths, batch_sizes):
            data_end_idx = data_start_idx + dataset_length
            start_idx = data_start_idx + round(dataset_length / num_replicas * rank)
            end_idx = data_start_idx + round(dataset_length / num_replicas * (rank + 1))
            subset_sampler = BatchSampler(
                SubsetRandomSampler(list(range(start_idx, end_idx)), generator),
                batch_size,
                True,
            )
            self.subset_samplers.append(subset_sampler)
            num_batches.append((end_idx - start_idx) // batch_size)
            data_start_idx = data_end_idx
        self.length = (
            min(num_batches)
            if self.complete_dataset_iteration is None
            else num_batches[self.complete_dataset_iteration]
        )

    def __iter__(self) -> Iterator[List[int]]:
        """Creates a batch with groups of indices where each group corresponds to one dataset."""
        if self.complete_dataset_iteration is None:
            # Default case is iterating once over the shortest iterator. Works nicely with zip.
            for samples in zip(*self.subset_samplers):
                yield [j for i in samples for j in i]
        else:
            # Iterating over a specific iterator requires dealing with the length of other
            # iterators.
            iterators = [iter(sampler) for sampler in self.subset_samplers]
            for s in iterators[self.complete_dataset_iteration]:
                samples = []
                for i, iterator in enumerate(iterators):
                    if i != self.complete_dataset_iteration:
                        try:
                            samples += next(iterator)
                        except StopIteration:
                            iterators[i] = iter(self.subset_samplers[i])
                            samples += next(iterators[i])
                    else:
                        samples += s
                yield samples

    def __len__(self):
        return self.length



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.defaults import TASK_ID
from renate.utils.optimizer import create_partial_optimizer


@pytest.mark.parametrize(
    "optimizer,kwargs",
    [
        ("SGD", {"lr": 0.01, "momentum": 0.0, "weight_decay": 0.5}),
        ("Adam", {"lr": 0.01, "weight_decay": 0.5}),
    ],
)
def test_make_optimizer_with_different_configurations(optimizer, kwargs):
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    params = model.get_params(task_id=TASK_ID)

    opt = create_partial_optimizer(optimizer=optimizer, **kwargs)(params)
    assert isinstance(opt, torch.optim.Optimizer)


def test_unknown_optimizer_raises_error():
    optimizer_name = "Unknown Optimizer"
    with pytest.raises(ValueError, match=f"Unknown optimizer: {optimizer_name}."):
        create_partial_optimizer(optimizer=optimizer_name, lr=0.01, momentum=0.0, weight_decay=0.5)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest

from renate.utils.deepspeed import convert_to_tensor, recover_object_from_tensor


@pytest.mark.parametrize(
    "obj",
    [
        {
            "constructor_arguments": {
                "a": 1,
                "b": "2",
                "nested_dict": {
                    "k1": "v1",
                    "k2": "v2",
                },
            },
            "tasks_params_ids": "task_param_id_1",
            "misc_args": tuple(range(10)),
        }
    ],
)
def test_serialize_random_objects(obj):
    assert recover_object_from_tensor(convert_to_tensor(obj)) == obj



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os

import numpy as np
import pytest
import torch
import torchvision.transforms as transforms
from PIL import Image
from torch.utils.data import TensorDataset

from renate.data import ImageDataset
from renate.data.datasets import IndexedSubsetDataset, _EnumeratedDataset, _TransformedDataset


class MulTransform:
    def __init__(self, factor=3):
        self.factor = factor

    def __call__(self, x):
        return x * self.factor


@pytest.mark.parametrize("transform", [lambda x: x + 1, torch.sqrt, None])
@pytest.mark.parametrize("target_transform", [lambda x: x * 2, lambda x: x**2, None])
def test_transformed_dataset(transform, target_transform):
    X = torch.arange(10)
    y = torch.arange(10)
    ds = TensorDataset(X, y)
    ds_transformed = _TransformedDataset(ds, transform, target_transform)
    X_transformed = torch.stack([ds_transformed[i][0] for i in range(len(ds_transformed))], dim=0)
    y_transformed = torch.stack([ds_transformed[i][1] for i in range(len(ds_transformed))], dim=0)
    X_transformed_exp = X if transform is None else transform(X)
    y_transformed_exp = y if target_transform is None else target_transform(y)
    assert torch.equal(X_transformed, X_transformed_exp)
    assert torch.equal(y_transformed, y_transformed_exp)


def test_enumerated_dataset():
    ds = TensorDataset(torch.arange(10) ** 2)
    ds_enumerated = _EnumeratedDataset(ds)
    for i in range(10):
        idx, batch = ds_enumerated[i]
        assert i == idx
        assert batch[0] == idx**2


def test_image_dataset(tmpdir):
    image = np.ones((10, 10, 3), dtype=np.uint8) * 255
    image = Image.fromarray(image)
    image.save(os.path.join(tmpdir, "image.png"))
    dataset = ImageDataset(
        [os.path.join(tmpdir, "image.png")],
        [3],
        transform=transforms.Compose([transforms.ToTensor(), MulTransform(2)]),
        target_transform=MulTransform(3),
    )
    data, target = dataset[0]
    assert torch.equal(data, torch.ones(3, 10, 10) * 2)
    assert torch.equal(target, torch.tensor(3 * 3))


@pytest.mark.parametrize("indexes_to_keep", [1, 0, [0, 1], (0, 1)])
def test_indexed_dataset(indexes_to_keep):
    X = [torch.arange(10), torch.arange(20)[::2]]
    ds = TensorDataset(*X)
    subset = IndexedSubsetDataset(ds, indexes_to_keep=indexes_to_keep)
    if isinstance(indexes_to_keep, (list, tuple)):
        curr_x = torch.vstack([X[ind] for ind in indexes_to_keep]).T
    else:
        curr_x = X[indexes_to_keep]
    ds_elements = torch.tensor([subset[i] for i in range(len(subset))])
    assert torch.equal(curr_x, ds_elements)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Callable, List

import torch
from torch.nn import Parameter
from torch.optim import Optimizer

import renate.defaults as defaults


def create_partial_optimizer(
    optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
    lr: float = defaults.LEARNING_RATE,
    momentum: float = defaults.MOMENTUM,
    weight_decay: float = defaults.WEIGHT_DECAY,
) -> Callable[[List[Parameter]], Optimizer]:
    """Creates a partial optimizer object.

    Args:
        optimizer: The name of the optimizer to be used. Options: `Adam` or `SGD`.
        lr: Learning rate to be used.
        momentum: Value for the momentum hyperparameter (if relevant).
        weight_decay: Value for the weight_decay hyperparameter (if relevant).
    """

    if optimizer == "SGD":
        return partial(torch.optim.SGD, lr=lr, momentum=momentum, weight_decay=weight_decay)
    elif optimizer == "Adam":
        return partial(torch.optim.Adam, lr=lr, weight_decay=weight_decay)
    else:
        raise ValueError(f"Unknown optimizer: {optimizer}.")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import shutil
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd
import torch
from pytorch_lightning import seed_everything

import renate
import renate.defaults as defaults
from renate.cli.parsing_functions import (
    get_data_module_fn_kwargs,
    get_metrics_fn_kwargs,
    get_model_fn_kwargs,
    get_scheduler_kwargs,
    get_transforms_kwargs,
)
from renate.evaluation.metrics.classification import (
    average_accuracy,
    backward_transfer,
    forgetting,
    forward_transfer,
    micro_average_accuracy,
)
from renate.training import run_training_job
from renate.training.training import submit_remote_job
from renate.utils.file import (
    copy_to_uri,
    is_s3_uri,
    move_to_uri,
    save_pandas_df_to_csv,
)
from renate.utils.module import (
    evaluate_and_record_results,
    get_and_prepare_data_module,
    get_metrics,
    get_model,
    import_module,
)

logger = logging.getLogger(__name__)


def experiment_config_file():
    return str(Path(renate.__path__[0]) / "benchmark" / "experiment_config.py")


def create_cumulative_metrics() -> List[Tuple[str, Callable]]:
    """Gets the cumulative metrics for a given task along with a name of the metric to include in
    any potential results table.
    """
    return [
        ("Average Accuracy", average_accuracy),
        ("Micro Average Accuracy", micro_average_accuracy),
        ("Forgetting", forgetting),
        ("Forward Transfer", forward_transfer),
        ("Backward Transfer", backward_transfer),
    ]


def cumulative_metrics_summary(
    results: Dict[str, List[List[float]]],
    cumulative_metrics: List[Tuple[str, Callable]],
    num_tasks: int,
    num_instances: List[int],
) -> pd.DataFrame:
    """Creates a pandas DataFrame summary with respect to the observed tasks, specified by
    `num_tasks`.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        cumulative_metrics: The list of (name, metric) tuples.
        num_tasks: The total number of tasks.
        num_instances: Count of test data points for each task.
    """
    data = []
    for task_id in range(num_tasks):
        row = [task_id + 1]
        for _, metric in cumulative_metrics:
            row.append(metric(results, task_id, num_instances))
        data.append(row)

    column_names = ["Task ID"] + [name for name, _ in cumulative_metrics]
    df = pd.DataFrame(data, columns=column_names)
    return df


def individual_metrics_summary(
    results: Dict[str, List[List[float]]],
    current_task: int,
    num_tasks: int,
) -> pd.DataFrame:
    """Creates a pandas DataFrame summary for all individual metrics with respect to all observed
    tasks.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        current_task: The current task ID.
        num_tasks: The total number of tasks.
    """
    data = []
    metric_columns = [k for k in results.keys() if "_init" not in k]

    for task_id in range(current_task):
        row = [task_id + 1]
        for key in metric_columns:
            value = results[key]
            for v in value[task_id]:
                row.append(v)
        data.append(row)

    sub_columns = [f"Task {i}" for i in range(1, num_tasks + 1)]
    mux = pd.MultiIndex.from_product([metric_columns, sub_columns])
    mux = mux.insert(0, "Task ID")
    df = pd.DataFrame(data, columns=mux)
    return df


def execute_experiment_job(
    backend: defaults.SUPPORTED_BACKEND_TYPE,
    config_file: str,
    config_space: Dict[str, Any],
    experiment_outputs_url: str,
    mode: defaults.SUPPORTED_TUNING_MODE_TYPE,
    metric: str,
    num_updates: int,
    working_directory: Optional[str] = defaults.WORKING_DIRECTORY,
    requirements_file: Optional[str] = None,
    dependencies: Optional[List[str]] = None,
    role: Optional[str] = None,
    instance_type: str = defaults.INSTANCE_TYPE,
    instance_count: int = defaults.INSTANCE_COUNT,
    instance_max_time: float = defaults.INSTANCE_MAX_TIME,
    max_time: Optional[float] = None,
    max_num_trials_started: Optional[int] = None,
    max_num_trials_completed: Optional[int] = None,
    max_num_trials_finished: Optional[int] = None,
    n_workers: int = defaults.N_WORKERS,
    seed: int = defaults.SEED,
    accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
    devices: int = defaults.DEVICES,
    deterministic_trainer: bool = True,
    gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
    gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
    job_name: str = defaults.JOB_NAME,
    strategy: str = defaults.DISTRIBUTED_STRATEGY,
    precision: str = defaults.PRECISION,
    save_state: bool = defaults.SAVE_BENCHMARK_STATE,
) -> None:
    """Executes the experiment job.

    Args:
        backend: Backend of the experiment job.
        config_file: Path to the Renate config file.
        config_space: Details for defining your own search space is provided in the
            `Syne Tune Documentation
            <https://github.com/awslabs/syne-tune/blob/main/docs/search_space.md>`_.
        experiment_outputs_url: Path to the experiment outputs.
        mode: Whether to minimize or maximize the metric.
        metric: Metric of the experiment job.
        num_updates: Number of updates of the experiment job.
        working_directory: Path to the working directory.
        requirements_file: Path to the requirements file.
        dependencies: (SageMaker backend only) List of strings containing absolute or relative paths
            to files and directories that will be uploaded as part of the SageMaker training job.
        role: Role of the experiment job.
        instance_type: Instance type of the experiment job.
        instance_count: Instance count of the experiment job.
        instance_max_time: Instance max time of the experiment job.
        max_time: Max time of the experiment job.
        max_num_trials_started: Max number of trials started of the experiment job.
        max_num_trials_completed: Max number of trials completed of the experiment job.
        max_num_trials_finished: Max number of trials finished of the experiment job.
        n_workers: Number of workers of the experiment job.
        seed: Seed of the experiment job.
        accelerator: Type of accelerator to use.
        devices: Number of devices to use.
        deterministic_trainer: When true the Trainer adopts a deterministic behaviour also on GPU.
            In this function this parameter is set to True by default.
        gradient_clip_val: The value at which to clip gradients. Passing None disables it.
            `More details <https://lightning.ai/docs/pytorch/stable/common/trainer.html#init>`__
        gradient_clip_algorithm: The gradient clipping algorithm to use. Can be norm or value.
            `More details <https://lightning.ai/docs/pytorch/stable/common/trainer.html#init>`__
        job_name: Name of the experiment job.
        strategy: Name of the distributed training strategy to use.
            `More details <https://lightning.ai/docs/pytorch/stable/extensions/strategy.html>`__
        precision: Type of bit precision to use.
            `More details <https://lightning.ai/docs/pytorch/stable/common/precision_basic.html>`__
        save_state: Flag to retain models and buffer states of each update step. Disable to save
            storage.
    """
    assert (
        mode in defaults.SUPPORTED_TUNING_MODE
    ), f"Mode {mode} is not in {defaults.SUPPORTED_TUNING_MODE}."
    assert (
        backend in defaults.SUPPORTED_BACKEND
    ), f"Backend {backend} is not in {defaults.SUPPORTED_BACKEND}."
    if backend == "local":
        return _execute_experiment_job_locally(
            config_file=config_file,
            experiment_outputs_url=experiment_outputs_url,
            mode=mode,
            config_space=config_space,
            metric=metric,
            working_directory=working_directory,
            num_updates=num_updates,
            max_time=max_time,
            max_num_trials_started=max_num_trials_started,
            max_num_trials_completed=max_num_trials_completed,
            max_num_trials_finished=max_num_trials_finished,
            n_workers=n_workers,
            accelerator=accelerator,
            devices=devices,
            deterministic_trainer=deterministic_trainer,
            seed=seed,
            strategy=strategy,
            precision=precision,
            save_state=save_state,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
        )
    _execute_experiment_job_remotely(
        job_name=job_name,
        config_file=config_file,
        experiment_outputs_url=experiment_outputs_url,
        mode=mode,
        metric=metric,
        num_updates=num_updates,
        working_directory=working_directory,
        dependencies=dependencies or [],
        config_space=config_space,
        max_time=max_time,
        max_num_trials_started=max_num_trials_started,
        max_num_trials_completed=max_num_trials_completed,
        max_num_trials_finished=max_num_trials_finished,
        n_workers=n_workers,
        accelerator=accelerator,
        devices=devices,
        deterministic_trainer=deterministic_trainer,
        gradient_clip_val=gradient_clip_val,
        gradient_clip_algorithm=gradient_clip_algorithm,
        seed=seed,
        requirements_file=requirements_file,
        role=role,
        instance_type=instance_type,
        instance_count=instance_count,
        instance_max_time=instance_max_time,
        strategy=strategy,
        precision=precision,
        save_state=save_state,
    )


def _execute_experiment_job_locally(
    config_file: str,
    experiment_outputs_url: str,
    num_updates: int,
    mode: defaults.SUPPORTED_TUNING_MODE_TYPE,
    config_space: Dict[str, Any],
    metric: str,
    working_directory: str,
    seed: int,
    accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE,
    devices: int,
    max_time: float,
    max_num_trials_started: int,
    max_num_trials_completed: int,
    max_num_trials_finished: int,
    n_workers: int,
    deterministic_trainer: bool,
    strategy: str,
    precision: str,
    save_state: bool,
    gradient_clip_val: Optional[float],
    gradient_clip_algorithm: Optional[str],
) -> None:
    """Runs an experiment, combining hyperparameter tuning and model for multiple updates.

    See renate.benchmark.experimentation.execute_experiment_job for more details.
    """
    logger.info("Start experiment.")
    seed_everything(seed, True)

    input_state_url = defaults.input_state_folder(working_directory)
    output_state_url = defaults.output_state_folder(working_directory)
    data_url = defaults.data_folder(working_directory)
    model_url = defaults.model_file(input_state_url)
    logs_url = defaults.logs_folder(working_directory)

    for url in [input_state_url, output_state_url, logs_url]:
        if os.path.exists(url):
            shutil.rmtree(url)
        Path(url).mkdir(parents=True, exist_ok=True)

    config_module = import_module("config_module", config_file)
    scheduler, scheduler_kwargs = get_scheduler_kwargs(config_module)
    model_fn_kwargs = get_model_fn_kwargs(config_module, config_space)
    logger.info(f"Loading model {model_fn_kwargs.get('model_name', '')}")
    model = get_model(config_module, **model_fn_kwargs)
    data_module_fn_kwargs = get_data_module_fn_kwargs(config_module, config_space)
    logger.info(f"Prepare dataset {data_module_fn_kwargs.get('dataset_name', '')}")
    data_module = get_and_prepare_data_module(
        config_module,
        data_path=data_url,
        chunk_id=0,
        seed=seed,
        **data_module_fn_kwargs,
    )
    data_module.setup()
    assert num_updates == len(
        data_module.test_data()
    ), f"The dataset has {len(data_module.test_data())} chunks, expected {num_updates}."
    num_instances = [len(data_chunk) for data_chunk in data_module.test_data()]
    transforms = get_transforms_kwargs(config_module, config_space)
    metrics_fn_kwargs = get_metrics_fn_kwargs(config_module, config_space)
    metrics = get_metrics(config_module, **metrics_fn_kwargs)

    torch.save(
        model.state_dict(),
        model_url,
    )

    # TODO: evaluate's trainer has to use devices=1:
    # See https://github.com/Lightning-AI/lightning/issues/2537
    # The fix is to launch evaluation in a separate process like training.
    results = evaluate_and_record_results(
        {},
        model=model,
        data_module=data_module,
        transform=transforms.get("test_transform"),
        target_transform=transforms.get("target_test_transform"),
        logged_metrics=metrics,
        metric_postfix="_init",
        accelerator=accelerator,
        devices=1,
        strategy=strategy,
        precision=precision,
    )

    for update_id in range(num_updates):
        logger.info(f"Starting Update {update_id + 1}/{num_updates}.")
        update_url = os.path.join(experiment_outputs_url, f"update_{update_id}")
        run_training_job(
            mode=mode,
            config_space=config_space,
            metric=metric,
            backend="local",
            updater=config_space["updater"],
            max_epochs=config_space["max_epochs"],
            chunk_id=update_id,
            input_state_url=input_state_url,
            output_state_url=output_state_url,
            working_directory=working_directory,
            config_file=config_file,
            max_time=max_time,
            max_num_trials_started=max_num_trials_started,
            max_num_trials_completed=max_num_trials_completed,
            max_num_trials_finished=max_num_trials_finished,
            n_workers=n_workers,
            scheduler=scheduler,
            scheduler_kwargs=scheduler_kwargs,
            seed=seed,
            accelerator=accelerator,
            devices=devices,
            precision=precision,
            strategy=strategy,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
        )
        move_to_uri(output_state_url, input_state_url)
        if save_state:
            copy_to_uri(input_state_url, update_url)
        model = get_model(
            config_module,
            model_state_url=model_url,
            **get_model_fn_kwargs(config_module, config_space),
        )

        results = evaluate_and_record_results(
            results,
            model=model,
            data_module=data_module,
            transform=transforms.get("test_transform"),
            target_transform=transforms.get("target_test_transform"),
            logged_metrics=metrics,
            accelerator=accelerator,
            devices=1,
        )
        df = individual_metrics_summary(results, update_id + 1, num_updates)
        save_pandas_df_to_csv(
            df, defaults.metric_summary_file(logs_url, special_str=f"_update_{update_id}")
        )
        logger.info(f"### Results after update {update_id + 1}: ###")
        logger.info(df)

    cumulative_metrics = create_cumulative_metrics()
    df = cumulative_metrics_summary(results, cumulative_metrics, num_updates, num_instances)
    save_pandas_df_to_csv(df, defaults.metric_summary_file(logs_url))
    logger.info("### Cumulative results: ###")
    logger.info(df)

    if not save_state:
        move_to_uri(defaults.hpo_file(input_state_url), str(experiment_outputs_url))
    move_to_uri(logs_url, defaults.logs_folder(experiment_outputs_url))

    shutil.rmtree(working_directory)
    logger.info("Experiment completed successfully.")


def _execute_experiment_job_remotely(experiment_outputs_url: str, **job_kwargs: Any) -> str:
    """Executes the experiment job on SageMaker.

    See renate.benchmark.experimentation.execute_experiment_job for more details.
    """
    assert is_s3_uri(
        experiment_outputs_url
    ), f"experiment_outputs_url {experiment_outputs_url} is not on S3."
    return submit_remote_job(
        experiment_outputs_url=experiment_outputs_url,
        optional_dependencies="benchmark",
        **job_kwargs,
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch
from scipy.stats import kstest

from renate.shift.detector import ShiftDetectorWithFeatureExtractor


class KolmogorovSmirnovCovariateShiftDetector(ShiftDetectorWithFeatureExtractor):
    """A Kolmogorov-Smirnov (KS) test on each feature.

    A KS test is a univariate two-sample test, which we perform separately for each feature. To
    aggregate these tests without running into multiple-testing problems, we use a Bonferroni
    correction, as proposed in

        Rabanser et al. Failing Loudly: An Empirical Study of Methods for Detecting Dataset Shift.
        NeurIPS 2019.
    """

    def _fit_with_features(self, X: torch.Tensor) -> None:
        self._X_ref = X

    def _score_with_features(self, X: torch.Tensor) -> float:
        n_features = X.size(1)
        p_vals = [
            kstest(X[:, i].numpy(), self._X_ref[:, i].numpy()).pvalue for i in range(n_features)
        ]
        # Bonferroni correction: Reject only if the minimal p-value among the multiple tests is
        # lower than `alpha / num_tests`, where `alpha` is the significance level. Equivalently, we
        # multiply the p-value by `num_tests`.
        p_val = min(1.0, min(p_vals) * n_features)
        return 1.0 - p_val



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

from renate import defaults
from renate.updaters.avalanche.plugins import RenateFileSystemCheckpointStorage


def test_renate_file_system_checkpoint_storage_plugin(tmpdir):
    expected_directory = tmpdir
    checkpoint_name = "checkpoint_name"
    plugin = RenateFileSystemCheckpointStorage(directory=expected_directory)
    assert plugin._make_checkpoint_dir(checkpoint_name) == expected_directory
    assert defaults.learner_state_file(expected_directory) == str(
        plugin._make_checkpoint_file_path(checkpoint_name)
    )
    assert not plugin.checkpoint_exists(checkpoint_name)
    Path(defaults.learner_state_file(expected_directory)).touch()
    assert plugin.checkpoint_exists(checkpoint_name)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch.nn as nn


class SharedMultipleLinear(nn.ModuleDict):
    """This implements a linear classification layer for multiple tasks (updates).
    This linear layer can be shared across all tasks or can have a separate layer per task.
    This follows the `_task_params` in the `RenateBenchmarkingModule` that is a `nn.ModuleDict`
    that holds a classifier per task (as in TIL).

    Args:
        in_features: size of each input sample
        out_features: size of each output sample
        bias: If set to ``False``, the layer will not learn an additive bias.
            Default: ``True``
        share_parameters: Flag whether to share parameters or use individual linears per task.
            The interface remains identical, and the underlying linear layer is shared (or not).
        num_updates: Number of updates that have happened/is happening.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        share_parameters: bool = True,
        num_updates: int = 0,
    ) -> None:
        self._share_parameters = share_parameters
        self.in_features = in_features
        self.out_features = out_features
        self.bias = bias

        super().__init__()
        for _ in range(num_updates):
            self.increment_task()

    def increment_task(self) -> None:
        currlen = len(self)
        if self._share_parameters:
            self[f"{currlen}"] = (
                self[list(self.keys())[0]]
                if currlen > 0
                else nn.Linear(in_features=self.in_features, out_features=self.out_features)
            )
        else:
            self[f"{currlen}"] = nn.Linear(
                in_features=self.in_features, out_features=self.out_features
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, Tuple, Union

import torch

NestedTensors = Union[torch.Tensor, Tuple["NestedTensors"], Dict[str, "NestedTensors"]]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import inspect
import sys
from pathlib import Path
from typing import Dict, List, Optional, Union

import pytest

from renate.cli.parsing_functions import (
    CUSTOM_ARGS_GROUP,
    REQUIRED_ARGS_GROUP,
    get_argument_type,
    get_data_module_fn_kwargs,
    get_function_args,
    get_metrics_fn_kwargs,
    get_model_fn_kwargs,
    to_dense_str,
)
from renate.utils.module import import_module

config_file = str(Path(__file__).parent.parent / "renate_config_files" / "config.py")
config_module = import_module("config_module", config_file)


def test_get_argument_type():
    """Test if correct argument type is extracted from typed function."""

    def foo(
        int_param: int,
        float_list_param: List[float],
        tuple_param: tuple,
        dict_param: Dict[str, int],
        union_param: Union[str, int],
        list_of_float_lists: List[List[float]],
        no_annotation_param,
        optional_str_param: Optional[str] = None,
        optional_dict_param: Optional[Dict[str, str]] = None,
    ):
        pass

    arg_spec = inspect.getfullargspec(foo)
    expected_types = {
        "int_param": int,
        "float_list_param": list,
        "tuple_param": tuple,
        "optional_str_param": str,
        "list_of_float_lists": list,
    }
    expected_errors = {
        "dict_param": r"Type typing.Dict\[str, int\] is not supported \(argument dict_param\).",
        "union_param": r"Type typing.Union\[str, int\] is not supported \(argument union_param\).",
        "optional_dict_param": r"Type typing.Union\[typing.Dict\[str, str\], NoneType\] is not "
        r"supported \(argument optional_dict_param\).",
        "no_annotation_param": r"Missing type annotation for argument no_annotation_param.",
    }
    if sys.version_info.minor >= 9:
        expected_errors["optional_dict_param"] = (
            r"Type typing.Optional\[typing.Dict\[str, str\]\] is not supported "
            r"\(argument optional_dict_param\)."
        )
    for argument_name, expected_type in expected_types.items():
        assert get_argument_type(arg_spec=arg_spec, argument_name=argument_name) == expected_type

    for argument_name, expected_error in expected_errors.items():
        with pytest.raises(TypeError, match=expected_error):
            get_argument_type(arg_spec=arg_spec, argument_name=argument_name)


@pytest.mark.parametrize(
    "all_args,ignore_args",
    [
        ({}, []),
        ({}, ["val_size", "seed"]),
        (
            {
                "existing_arg": {
                    "type": float,
                    "argument_group": REQUIRED_ARGS_GROUP,
                    "required": True,
                }
            },
            ["data_path"],
        ),
    ],
    ids=("No all_args, no ignore_args", "no all_args", "all_args"),
)
def test_get_function_args(all_args, ignore_args):
    expected_args = [
        "data_path",
        "val_size",
        "seed",
        "groupings",
        "optional_tuple",
        "optional_float",
        "list_param",
        "bool_param",
    ]
    expected_all_args = {
        **all_args,
        **{
            "data_path": {
                "type": str,
                "argument_group": CUSTOM_ARGS_GROUP,
                "required": True,
                "true_type": str,
            },
            "val_size": {
                "type": float,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": 0.0,
                "true_type": float,
            },
            "seed": {
                "type": int,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": 0,
                "true_type": int,
            },
            "groupings": {
                "type": str,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": "((0,1),(2,3,4))",
                "true_type": tuple,
            },
            "optional_tuple": {
                "type": str,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": None,
                "true_type": tuple,
            },
            "optional_float": {
                "type": float,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": None,
                "true_type": float,
            },
            "list_param": {
                "type": str,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": "[1,2]",
                "true_type": list,
            },
            "bool_param": {
                "type": str,
                "argument_group": CUSTOM_ARGS_GROUP,
                "default": "False",
                "true_type": bool,
            },
        },
    }
    for arg in ignore_args:
        del expected_all_args[arg]
    args = get_function_args(
        config_module=config_module,
        function_name="data_module_fn",
        all_args=all_args,
        ignore_args=ignore_args,
    )
    assert args == expected_args
    assert all_args == expected_all_args


def test_get_function_args_with_inconsistent_args():
    """If types of arguments are different, an error should be raised."""
    all_args = {
        "model_state_url": {
            "type": float,
            "argument_group": REQUIRED_ARGS_GROUP,
            "required": True,
        }
    }

    with pytest.raises(
        TypeError,
        match=r"Types of `model_state_url` are not consistent. Defined as type `<class 'str'>` "
        r"as well as `<class 'float'>`.",
    ):
        get_function_args(
            config_module=config_module,
            function_name="model_fn",
            all_args=all_args,
            ignore_args=[],
        )


def test_get_function_args_prefers_required_true():
    """If argument is defined multiple times, it is required if it is required only once."""
    all_args = {
        "data_path": {
            "type": str,
            "argument_group": REQUIRED_ARGS_GROUP,
            "required": False,
        }
    }
    get_function_args(
        config_module=config_module,
        function_name="data_module_fn",
        all_args=all_args,
        ignore_args=[],
    )
    assert all_args["data_path"]["required"]


def test_get_fn_kwargs_helper_functions():
    """Tests whether the different helper functions correctly create kwargs given a dictionary
    and the Python function."""
    expected_data_module_kwargs = {
        "data_path": "home/data/path",
        "groupings": ((1, 2), (3, 4)),
        "optional_float": None,
        "bool_param": False,
    }
    config_space = {
        "data_path": expected_data_module_kwargs["data_path"],
        "model_state_url": "home/model/state",
        "unused_config": 1,
        "groupings": to_dense_str(expected_data_module_kwargs["groupings"]),
        "optional_float": to_dense_str(expected_data_module_kwargs["optional_float"]),
        "bool_param": to_dense_str(expected_data_module_kwargs["bool_param"]),
        "num_outputs": 10,
    }
    data_module_kwargs = get_data_module_fn_kwargs(
        config_module=config_module, config_space=config_space, cast_arguments=True
    )
    assert data_module_kwargs == expected_data_module_kwargs
    model_kwargs = get_model_fn_kwargs(
        config_module=config_module, config_space=config_space, cast_arguments=True
    )
    assert model_kwargs == {"model_state_url": config_space["model_state_url"]}
    metrics_kwargs = get_metrics_fn_kwargs(
        config_module=config_module, config_space=config_space, cast_arguments=True
    )
    assert metrics_kwargs == {"num_outputs": config_space["num_outputs"]}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch
from sklearn.cluster import KMeans

from renate.models.task_identification_strategies import TaskPrototypes


def test_task_prototypes():
    data = torch.nn.functional.normalize(torch.rand(10, 3))
    labels = torch.arange(start=0, end=data.size(0))
    task_proto = TaskPrototypes(0, 0, data.size(1))
    # lets attach
    task_proto._training_feat_centroids = data
    task_proto._training_feat_task_ids = labels

    test_data = torch.nn.functional.normalize(torch.rand(5, 3))
    predictions = task_proto.infer_task(test_data)

    kmeans = KMeans(n_clusters=data.size(0))
    kmeans.cluster_centers_ = data.numpy()
    kmeans.labels_ = labels.numpy()
    kmeans._n_threads = 1

    gnd_truth = kmeans.predict(test_data.numpy())

    assert (predictions.numpy() == gnd_truth).all()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
from typing import Any, Dict, List, Optional, Tuple

import torch

from renate.models import RenateModule
from renate.types import NestedTensors


class Component(abc.ABC):
    """The abstract class implementing a Component, usable in the BaseExperienceReplayLearner.

    This is an abstract class from which each other component e.g. additional
    regularising loss or a module updater should inherit from.
    The components should be a modular and independent to an extent where they can be composed
    together in an ordered list to be deployed in the BaseExperienceReplayLearner.

    Args:
        weight: A scaling coefficient which should scale the loss which gets returned.
        sample_new_memory_batch: Whether a new batch of data should be sampled from the memory
            buffer when the loss is calculated.
    """

    def __init__(self, weight: float = 0, sample_new_memory_batch: bool = False) -> None:
        self.weight = weight
        self.sample_new_memory_batch = sample_new_memory_batch
        self._verify_attributes()

    def loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: Optional[List[torch.Tensor]],
    ) -> torch.Tensor:
        """Computes some user-defined loss which is added to the main training loss in the training
        step.

        Args:
            outputs_memory: The outputs of the model with respect to memory data (batch_memory).
            batch_memory: The batch of data sampled from the memory buffer, including the meta data.
            intermediate_representation_memory: Intermediate feature representations of the network
                upon passing the input through the network.
        """
        return torch.tensor(0.0)

    def on_train_start(self, model: RenateModule) -> None:
        """Updates the model parameters.

        Args:
            model: The model used for training.
        """
        pass

    def on_train_batch_end(self, model: RenateModule) -> None:
        """Internally records a training and optimizer step in the component.

        Args:
            model: The model used for training.
        """
        pass

    def _verify_attributes(self) -> None:
        """Verify if attributes have valid values."""
        pass

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Load relevant information from checkpoint."""
        pass

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Add relevant information to checkpoint."""
        pass



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from renate.data.data_module import CSVDataModule
from renate.data.datasets import ImageDataset

__all__ = ["CSVDataModule", "ImageDataset"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import argparse
import ast
import inspect
import sys
from importlib.util import find_spec
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import pytorch_lightning as pl
from syne_tune.optimizer.scheduler import TrialScheduler

from renate import defaults
from renate.updaters.experimental.er import (
    CLSExperienceReplayModelUpdater,
    DarkExperienceReplayModelUpdater,
    ExperienceReplayModelUpdater,
    PooledOutputDistillationExperienceReplayModelUpdater,
    SuperExperienceReplayModelUpdater,
)
from renate.updaters.experimental.fine_tuning import FineTuningModelUpdater
from renate.updaters.experimental.gdumb import GDumbModelUpdater
from renate.updaters.experimental.joint import JointModelUpdater
from renate.updaters.experimental.l2p import (
    LearningToPromptModelUpdater,
    LearningToPromptReplayModelUpdater,
)
from renate.updaters.experimental.offline_er import OfflineExperienceReplayModelUpdater
from renate.updaters.experimental.repeated_distill import RepeatedDistillationModelUpdater
from renate.updaters.experimental.speft import SPeftModelUpdater
from renate.updaters.model_updater import ModelUpdater

REQUIRED_ARGS_GROUP = "Required Arguments"
RENATE_STATE_ARGS_GROUP = "Renate State"
HYPERPARAMETER_ARGS_GROUP = "Hyperparameters"
CUSTOM_ARGS_GROUP = "Custom Arguments"
OPTIONAL_ARGS_GROUP = "Optional Arguments"
DO_NOT_CHANGE_GROUP = "DO NOT CHANGE"
ARGS_GROUP_ORDER = [
    REQUIRED_ARGS_GROUP,
    RENATE_STATE_ARGS_GROUP,
    HYPERPARAMETER_ARGS_GROUP,
    CUSTOM_ARGS_GROUP,
    OPTIONAL_ARGS_GROUP,
    DO_NOT_CHANGE_GROUP,
]


def get_updater_and_learner_kwargs(
    args: argparse.Namespace,
) -> Tuple[Type[ModelUpdater], Dict[str, Any]]:
    """Returns the model updater class and the keyword arguments for the learner."""
    if args.updater.startswith("Avalanche-") and find_spec("avalanche", None) is None:
        raise ImportError("Avalanche is not installed. Please run `pip install Renate[avalanche]`.")
    learner_args = ["batch_size", "seed", "mask_unused_classes"]
    base_er_args = learner_args + [
        "loss_weight",
        "ema_memory_update_gamma",
        "memory_size",
        "batch_memory_frac",
        "loss_normalization",
    ]
    updater_class = None
    if args.updater == "ER":
        learner_args = base_er_args + ["alpha"]
        updater_class = ExperienceReplayModelUpdater
    elif args.updater == "LearningToPrompt":
        learner_args = learner_args + ["prompt_sim_loss_weight"]
        updater_class = LearningToPromptModelUpdater
    elif args.updater == "LearningToPromptReplay":
        learner_args = learner_args + ["prompt_sim_loss_weight", "memory_size", "memory_batch_size"]
        updater_class = LearningToPromptReplayModelUpdater
    elif args.updater == "SPeft":
        updater_class = SPeftModelUpdater
    elif args.updater == "DER":
        learner_args = base_er_args + ["alpha", "beta"]
        updater_class = DarkExperienceReplayModelUpdater
    elif args.updater == "POD-ER":
        learner_args = base_er_args + ["alpha", "distillation_type", "normalize"]
        updater_class = PooledOutputDistillationExperienceReplayModelUpdater
    elif args.updater == "CLS-ER":
        learner_args = base_er_args + [
            "alpha",
            "beta",
            "stable_model_update_weight",
            "plastic_model_update_weight",
            "stable_model_update_probability",
            "plastic_model_update_probability",
        ]
        updater_class = CLSExperienceReplayModelUpdater
    elif args.updater == "Super-ER":
        learner_args = base_er_args + [
            "der_alpha",
            "der_beta",
            "sp_shrink_factor",
            "sp_sigma",
            "cls_alpha",
            "cls_stable_model_update_weight",
            "cls_plastic_model_update_weight",
            "cls_stable_model_update_probability",
            "cls_plastic_model_update_probability",
            "pod_alpha",
            "pod_distillation_type",
            "pod_normalize",
        ]
        updater_class = SuperExperienceReplayModelUpdater
    elif args.updater == "Offline-ER":
        learner_args = learner_args + ["loss_weight_new_data", "memory_size", "batch_memory_frac"]
        updater_class = OfflineExperienceReplayModelUpdater
    elif args.updater == "RD":
        learner_args = learner_args + ["memory_size"]
        updater_class = RepeatedDistillationModelUpdater
    elif args.updater == "GDumb":
        learner_args = learner_args + ["memory_size"]
        updater_class = GDumbModelUpdater
    elif args.updater == "Joint":
        learner_args = learner_args
        updater_class = JointModelUpdater
    elif args.updater == "FineTuning":
        learner_args = learner_args
        updater_class = FineTuningModelUpdater
    elif args.updater == "Avalanche-ER":
        learner_args = learner_args + ["memory_size", "batch_memory_frac"]
        from renate.updaters.avalanche.model_updater import ExperienceReplayAvalancheModelUpdater

        updater_class = ExperienceReplayAvalancheModelUpdater
    elif args.updater == "Avalanche-EWC":
        learner_args = learner_args + ["ewc_lambda"]
        from renate.updaters.avalanche.model_updater import ElasticWeightConsolidationModelUpdater

        updater_class = ElasticWeightConsolidationModelUpdater
    elif args.updater == "Avalanche-LwF":
        learner_args = learner_args + ["alpha", "temperature"]
        from renate.updaters.avalanche.model_updater import LearningWithoutForgettingModelUpdater

        updater_class = LearningWithoutForgettingModelUpdater
    elif args.updater == "Avalanche-iCaRL":
        learner_args = learner_args + ["memory_size", "batch_memory_frac"]
        from renate.updaters.avalanche.model_updater import ICaRLModelUpdater

        updater_class = ICaRLModelUpdater
    if updater_class is None:
        raise ValueError(f"Unknown learner {args.updater}.")
    learner_kwargs = {arg: value for arg, value in vars(args).items() if arg in learner_args}
    return updater_class, learner_kwargs


def _cast_arguments_to_true_types(
    arguments: Dict[str, Dict[str, Any]], args_dict: Dict[str, str]
) -> None:
    """Casts all values in args_dict to the value specified in arguments.

    Booleans, lists, None, and tuples are passed as a string to the script and argsparse will not
    convert them for us. After detecting the true types according to typing (typing is saved in
    ``arguments``), we now explicitly cast our arguments.
    After this step, ``args`` as a results of argparse.parse will contain the correct types and no
    further typecasting is required.
    """
    for argument_name, argument_kwargs in arguments.items():
        if args_dict[argument_name] == "None":
            args_dict[argument_name] = None
        if args_dict[argument_name] is None:
            continue
        if argument_kwargs.get("true_type") == bool:
            args_dict[argument_name] = args_dict[argument_name] == "True"
        elif argument_kwargs.get("true_type") in [list, tuple]:
            args_dict[argument_name] = ast.literal_eval(args_dict[argument_name])


def parse_arguments(
    config_module: ModuleType, function_names: List[str], ignore_args: List[str]
) -> Tuple[argparse.Namespace, Dict[str, Any]]:
    """Parses all input arguments.

    Combines standard arguments with custom arguments in functions of Renate config.

    Args:
        config_module: Module containing function of which we extract arguments.
        function_names: List of function names in module for which extract arguments.
        ignore_args: List of arguments to be ignored since they are not passed via CLI.
    Returns:
        First return value are the passed arguments where strings are converted to Booleans, lists
        and tuples. Second return value is a dictionary containing the arguments that will be passed
        to all functions specified in ``function_names``.
    """
    arguments = _standard_arguments()
    _add_hyperparameter_arguments(arguments, "optimizer_fn" not in vars(config_module))
    function_args = {}
    for function_name in function_names:
        function_args[function_name] = get_function_args(
            config_module=config_module,
            function_name=function_name,
            all_args=arguments,
            ignore_args=ignore_args,
        )

    parser = argparse.ArgumentParser()

    for argument_group_name in ARGS_GROUP_ORDER:
        argument_group = parser.add_argument_group(argument_group_name)
        for argument_name, argument_kwargs in arguments.items():
            if argument_kwargs["argument_group"] == argument_group_name:
                argument_group.add_argument(
                    f"--{argument_name}",
                    **{
                        key: value
                        for key, value in argument_kwargs.items()
                        if key != "argument_group" and key != "true_type"
                    },
                )
    args = parser.parse_args()
    _cast_arguments_to_true_types(arguments=arguments, args_dict=vars(args))
    return args, function_args


def _standard_arguments() -> Dict[str, Dict[str, Any]]:
    """Returns information about the minimum number of arguments accepted by ``run_training.py``."""
    return {
        "updater": {
            "type": str,
            "required": True,
            "choices": list(parse_by_updater),
            "help": "Select the type of model update strategy.",
            "argument_group": REQUIRED_ARGS_GROUP,
        },
        "config_file": {
            "type": str,
            "required": True,
            "help": "Location of python file containing model_fn and data_module_fn.",
            "argument_group": REQUIRED_ARGS_GROUP,
        },
        "input_state_url": {
            "type": str,
            "help": "Location of previous Renate state (if available).",
            "argument_group": RENATE_STATE_ARGS_GROUP,
        },
        "output_state_url": {
            "type": str,
            "help": "Location where to store the next Renate state.",
            "argument_group": RENATE_STATE_ARGS_GROUP,
        },
        "max_epochs": {
            "type": int,
            "default": defaults.MAX_EPOCHS,
            "help": "Maximum number of (finetuning-equivalent) epochs. "
            f"Default: {defaults.MAX_EPOCHS}",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "task_id": {
            "type": str,
            "default": defaults.TASK_ID,
            "help": "Task ID matching the current dataset. If you do not distinguish between "
            "different tasks, ignore this"
            f" argument. Default: {defaults.TASK_ID}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "metric": {
            "type": str,
            "help": "Metric monitored during training to save checkpoints.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "mode": {
            "type": str,
            "default": "min",
            "help": "Indicate whether a smaller `metric` is better (`min`) or a larger (`max`).",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "working_directory": {
            "type": str,
            "default": defaults.WORKING_DIRECTORY,
            "help": "Folder used by Renate to store files temporarily. Default: "
            f"{defaults.WORKING_DIRECTORY}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "seed": {
            "type": int,
            "default": defaults.SEED,
            "help": f"Seed used for this job. Default: {defaults.SEED}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "accelerator": {
            "type": str,
            "default": defaults.ACCELERATOR,
            "help": f"Accelerator used for this job. Default: {defaults.ACCELERATOR}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "devices": {
            "type": int,
            "default": defaults.DEVICES,
            "help": f"Devices used for this job. Default: {defaults.DEVICES} device.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "strategy": {
            "type": str,
            "default": defaults.DISTRIBUTED_STRATEGY,
            "help": "Distributed training strategy when devices > 1. Default:"
            + f"{defaults.DISTRIBUTED_STRATEGY}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
            "choices": list(pl.strategies.StrategyRegistry.keys()),
        },
        "precision": {
            "type": str,
            "default": defaults.PRECISION,
            "help": f"Distributed training precision. Default: {defaults.PRECISION}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
            "choices": ("16", "32", "64", "bf16"),
        },
        "early_stopping": {
            "type": str,
            "default": str(defaults.EARLY_STOPPING),
            "choices": ["True", "False"],
            "help": "Enables the early stopping of the optimization. Default: "
            f"{defaults.EARLY_STOPPING}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
            "true_type": bool,
        },
        "deterministic_trainer": {
            "type": str,
            "default": str(defaults.DETERMINISTIC_TRAINER),
            "choices": ["True", "False"],
            "help": "Enables deterministic training which may be slower. Default: "
            f"{defaults.DETERMINISTIC_TRAINER}.",
            "argument_group": OPTIONAL_ARGS_GROUP,
            "true_type": bool,
        },
        "gradient_clip_val": {
            "type": lambda x: None if x == "None" else float(x),
            "default": defaults.GRADIENT_CLIP_VAL,
            "help": "The value at which to clip gradients. None disables clipping.",
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "gradient_clip_algorithm": {
            "type": lambda x: None if x == "None" else x,
            "default": defaults.GRADIENT_CLIP_ALGORITHM,
            "help": "Gradient clipping algorithm to use.",
            "choices": ["norm", "value", None],
            "argument_group": OPTIONAL_ARGS_GROUP,
        },
        "mask_unused_classes": {
            "default": str(defaults.MASK_UNUSED_CLASSES),
            "type": str,
            "choices": ["True", "False"],
            "help": "Whether to use a class mask to kill the unused logits. Useful possibly for "
            "class incremental learning methods. ",
            "argument_group": OPTIONAL_ARGS_GROUP,
            "true_type": bool,
        },
        "prepare_data": {
            "type": str,
            "default": "True",
            "choices": ["True", "False"],
            "help": "Whether to call DataModule.prepare_data(). Default: True.",
            "argument_group": DO_NOT_CHANGE_GROUP,
            "true_type": bool,
        },
        "st_checkpoint_dir": {
            "type": str,
            "help": "Location for checkpoints.",
            "argument_group": DO_NOT_CHANGE_GROUP,
        },
    }


def _add_hyperparameter_arguments(
    arguments: Dict[str, Dict[str, Any]], add_optimizer_args: bool
) -> None:
    """Adds arguments for the specified updater."""
    updater: Optional[str] = None
    for i, arg in enumerate(sys.argv):
        if arg == "--updater" and len(sys.argv) > i:
            updater = sys.argv[i + 1]
            break
    if updater is None:
        return

    assert updater in parse_by_updater, f"Unknown updater {updater}."
    parse_by_updater[updater](arguments)
    _add_optimizer_arguments(arguments, add_optimizer_args)
    for value in arguments.values():
        if "argument_group" not in value:
            value["argument_group"] = HYPERPARAMETER_ARGS_GROUP


def _add_optimizer_arguments(
    arguments: Dict[str, Dict[str, Any]], add_optimizer_args: bool
) -> None:
    """A helper function that adds optimizer arguments."""
    if add_optimizer_args:
        arguments.update(
            {
                "optimizer": {
                    "type": str,
                    "default": defaults.OPTIMIZER,
                    "help": "Optimizer used for training. Options: SGD or Adam. Default: "
                    f"{defaults.OPTIMIZER}.",
                },
                "learning_rate": {
                    "type": float,
                    "default": defaults.LEARNING_RATE,
                    "help": "Learning rate used during model update. Default: "
                    f"{defaults.LEARNING_RATE}.",
                },
                "momentum": {
                    "type": float,
                    "default": defaults.MOMENTUM,
                    "help": f"Momentum used during model update. Default: {defaults.MOMENTUM}.",
                },
                "weight_decay": {
                    "type": float,
                    "default": defaults.WEIGHT_DECAY,
                    "help": "Weight decay used during model update. Default: "
                    f"{defaults.WEIGHT_DECAY}.",
                },
            }
        )
    arguments.update(
        {
            "batch_size": {
                "type": int,
                "default": defaults.BATCH_SIZE,
                "help": "Batch size used during model update for the new data. Default: "
                f"{defaults.BATCH_SIZE}.",
            },
            "loss_weight": {
                "type": float,
                "default": defaults.LOSS_WEIGHT,
                "help": f"Loss weight used during model update. Default: {defaults.LOSS_WEIGHT}.",
            },
        }
    )


def _add_replay_learner_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Replay Learner arguments."""
    arguments.update(
        {
            "memory_size": {
                "type": int,
                "default": defaults.MEMORY_SIZE,
                "help": "Memory size available for the memory buffer. Default: "
                f"{defaults.MEMORY_SIZE}.",
            },
            "batch_memory_frac": {
                "type": float,
                "default": defaults.BATCH_MEMORY_FRAC,
                "help": "Fraction of the batch populated with memory data. Default: "
                f"{defaults.BATCH_MEMORY_FRAC}.",
            },
        }
    )


def _add_base_experience_replay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Base Experience Replay arguments."""
    arguments.update(
        {
            "ema_memory_update_gamma": {
                "type": float,
                "default": defaults.EMA_MEMORY_UPDATE_GAMMA,
                "help": "Exponential moving average factor to update logits. Default: "
                f"{defaults.EMA_MEMORY_UPDATE_GAMMA}.",
            },
            "loss_normalization": {
                "type": int,
                "choices": [0, 1],
                "default": defaults.LOSS_NORMALIZATION,
                "help": "Whether to normalize the loss with respect to the loss weights. "
                f"Default: {bool(defaults.LOSS_NORMALIZATION)}.",
            },
        }
    )
    _add_replay_learner_arguments(arguments)


def _add_l2p_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    arguments.update(
        {
            "prompt_sim_loss_weight": {
                "type": float,
                "default": defaults.PROMPT_SIM_LOSS_WEIGHT,
                "help": "Prompt key similarity regularization weight. "
                f"Default: {defaults.PROMPT_SIM_LOSS_WEIGHT}",
            }
        }
    )


def _add_l2preplay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    _add_l2p_arguments(arguments)
    _add_offline_er_arguments(arguments)


def _add_speft_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    pass


def _add_gdumb_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds GDumb arguments."""
    _add_replay_learner_arguments(arguments)
    _add_joint_arguments(arguments)


def _add_joint_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Joint Learner arguments."""
    arguments.update(
        {
            "reset": {
                "type": str,
                "default": "True",
                "choices": ["True", "False"],
                "help": "Resets the model before the update. Default: True",
                "true_type": bool,
            },
        }
    )


def _add_finetuning_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Fine Tuning arguments."""
    pass


def _add_offline_er_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    _add_replay_learner_arguments(arguments)
    arguments.update(
        {
            "loss_weight_new_data": {
                "type": float,
                "default": None,
                "help": "Weight assigned to loss on the new data.",
            }
        }
    )


def _add_experience_replay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Experience Replay arguments."""
    arguments.update(
        {
            "alpha": {
                "type": float,
                "default": defaults.ER_ALPHA,
                "help": f"Weight for the loss of the buffer data. Default: {defaults.ER_ALPHA}.",
            }
        }
    )
    _add_base_experience_replay_arguments(arguments)


def _add_dark_experience_replay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Dark Experience Replay arguments."""
    arguments.update(
        {
            "alpha": {
                "type": float,
                "default": defaults.DER_ALPHA,
                "help": f"Weight for logit regularization term. Default: {defaults.DER_ALPHA}.",
            },
            "beta": {
                "type": float,
                "default": defaults.DER_BETA,
                "help": f"Weight for memory loss term. Default: {defaults.DER_BETA}.",
            },
        }
    )
    _add_base_experience_replay_arguments(arguments)


def _add_pod_experience_replay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds PODNet Experience Replay arguments."""
    arguments.update(
        {
            "alpha": {
                "type": float,
                "default": defaults.POD_ALPHA,
                "help": "Weight for intermediate representation regularization term. Default: "
                f"{defaults.POD_ALPHA}.",
            },
            "distillation_type": {
                "type": str,
                "default": defaults.POD_DISTILLATION_TYPE,
                "help": "Distillation type to apply with respect to the intermediate "
                f"representation. Default: {defaults.POD_DISTILLATION_TYPE}.",
            },
            "normalize": {
                "type": int,
                "default": defaults.POD_NORMALIZE,
                "help": "Whether to normalize both the current and cached features before computing"
                f" the Frobenius norm. Default: {defaults.POD_NORMALIZE}.",
            },
        }
    )
    _add_base_experience_replay_arguments(arguments)


def _add_cls_experience_replay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds CLS Experience Replay arguments."""
    arguments.update(
        {
            "alpha": {
                "type": float,
                "default": defaults.CLS_ALPHA,
                "help": f"Weight for the cross-entropy loss term. Default: {defaults.CLS_ALPHA}.",
            },
            "beta": {
                "type": float,
                "default": defaults.CLS_BETA,
                "help": "Weight for the consistency memory loss term. Default: "
                f"{defaults.CLS_BETA}.",
            },
            "stable_model_update_weight": {
                "type": float,
                "default": defaults.CLS_STABLE_MODEL_UPDATE_WEIGHT,
                "help": "The starting weight for the exponential moving average to update the "
                f"stable model. Default: {defaults.CLS_STABLE_MODEL_UPDATE_WEIGHT}.",
            },
            "plastic_model_update_weight": {
                "type": float,
                "default": defaults.CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
                "help": "The starting weight for the exponential moving average to update the "
                f"plastic model. Default: {defaults.CLS_PLASTIC_MODEL_UPDATE_WEIGHT}.",
            },
            "stable_model_update_probability": {
                "type": float,
                "default": defaults.CLS_STABLE_MODEL_UPDATE_PROBABILITY,
                "help": "Probability to update the stable model. Default: "
                f"{defaults.CLS_STABLE_MODEL_UPDATE_PROBABILITY}.",
            },
            "plastic_model_update_probability": {
                "type": float,
                "default": defaults.CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,
                "help": "Probability to update the plastic model. Default: "
                f"{defaults.CLS_PLASTIC_MODEL_UPDATE_PROBABILITY}.",
            },
        }
    )
    _add_base_experience_replay_arguments(arguments)


def _add_super_experience_replay_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Super Experience Replay arguments."""
    arguments.update(
        {
            "der_alpha": {
                "type": float,
                "default": defaults.SER_DER_ALPHA,
                "help": f"Weight for logit regularization term. Default: {defaults.SER_DER_ALPHA}.",
            },
            "der_beta": {
                "type": float,
                "default": defaults.SER_DER_BETA,
                "help": f"Weight for memory loss term. Default: {defaults.SER_DER_BETA}.",
            },
            "sp_shrink_factor": {
                "type": float,
                "default": defaults.SER_SP_SHRINK_FACTOR,
                "help": "Weight for logit regularization term. Default: "
                f"{defaults.SER_SP_SHRINK_FACTOR}.",
            },
            "sp_sigma": {
                "type": float,
                "default": defaults.SER_SP_SIGMA,
                "help": f"Weight for memory loss term. Default: {defaults.SER_SP_SIGMA}.",
            },
            "pod_alpha": {
                "type": float,
                "default": defaults.SER_POD_ALPHA,
                "help": "Weight for intermediate representation regularization term. Default: "
                f"{defaults.SER_POD_ALPHA}.",
            },
            "pod_distillation_type": {
                "type": str,
                "default": defaults.SER_POD_DISTILLATION_TYPE,
                "help": "Distillation type to apply with respect to the intermediate "
                f"intermediate representation. Default: {defaults.SER_POD_DISTILLATION_TYPE}.",
            },
            "pod_normalize": {
                "type": int,
                "default": defaults.SER_POD_NORMALIZE,
                "help": "Whether to normalize both the current and cached features before "
                f"computing the Frobenius norm. Default: {defaults.SER_POD_NORMALIZE}.",
            },
            "cls_alpha": {
                "type": float,
                "default": defaults.SER_CLS_ALPHA,
                "help": f"Weight for the consistency loss term. Default: {defaults.SER_CLS_ALPHA}.",
            },
            "cls_stable_model_update_weight": {
                "type": float,
                "default": defaults.SER_CLS_STABLE_MODEL_UPDATE_WEIGHT,
                "help": "The starting weight for the exponential moving average to update the "
                f"stable model. Default: {defaults.SER_CLS_STABLE_MODEL_UPDATE_WEIGHT}.",
            },
            "cls_plastic_model_update_weight": {
                "type": float,
                "default": defaults.SER_CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
                "help": "The starting weight for the exponential moving average to update the "
                f"plastic model. Default: {defaults.SER_CLS_PLASTIC_MODEL_UPDATE_WEIGHT}.",
            },
            "cls_stable_model_update_probability": {
                "type": float,
                "default": defaults.SER_CLS_STABLE_MODEL_UPDATE_PROBABILITY,
                "help": "Probability to update the stable model. Default: "
                f"{defaults.SER_CLS_STABLE_MODEL_UPDATE_PROBABILITY}.",
            },
            "cls_plastic_model_update_probability": {
                "type": float,
                "default": defaults.SER_CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,
                "help": "Probability to update the plastic model. Default: "
                f"{defaults.SER_CLS_PLASTIC_MODEL_UPDATE_PROBABILITY}.",
            },
        }
    )
    _add_base_experience_replay_arguments(arguments)


def _add_rd_learner_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds Repeated Distill Learner arguments."""
    arguments.update(
        {
            "memory_size": {
                "type": int,
                "default": defaults.MEMORY_SIZE,
                "help": f"Memory size available for the memory buffer. Default: "
                f"{defaults.MEMORY_SIZE}.",
            }
        }
    )


def _add_avalanche_ewc_learner_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds EWC arguments."""
    arguments.update(
        {
            "ewc_lambda": {
                "type": float,
                "default": defaults.EWC_LAMBDA,
                "help": f"EWC regularization hyperparameter. Default: {defaults.EWC_LAMBDA}.",
            }
        }
    )


def _add_avalanche_lwf_learner_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds LwF arguments."""
    arguments.update(
        {
            "alpha": {
                "type": float,
                "default": defaults.LWF_ALPHA,
                "help": f"Distillation loss weight. Default: {defaults.LWF_ALPHA}.",
            },
            "temperature": {
                "type": float,
                "default": defaults.LWF_TEMPERATURE,
                "help": "Temperature of the softmax function. Default: "
                f"{defaults.LWF_TEMPERATURE}.",
            },
        },
    )


def _add_avalanche_icarl_learner_arguments(arguments: Dict[str, Dict[str, Any]]) -> None:
    """A helper function that adds iCarl arguments."""
    arguments.update(
        {
            "memory_size": {
                "type": int,
                "default": defaults.MEMORY_SIZE,
                "help": f"Number of exemplars being stored. Default: {defaults.MEMORY_SIZE}.",
            }
        },
    )


def get_function_kwargs(args: argparse.Namespace, function_args: Dict[str, Any]) -> Dict[str, Any]:
    """Returns the kwargs for a function with defined arguments based on provided values.

    Args:
        args: Values for specific keys.
        function_args: Arguments of the function for which we extract values.
    """
    return {key: value for key, value in vars(args).items() if key in function_args}


def get_data_module_fn_kwargs(
    config_module, config_space: Dict[str, Any], cast_arguments: Optional[bool] = False
) -> Dict[str, Any]:
    """Returns the kwargs for a ``data_module_fn`` with defined arguments based on config_space."""
    return _get_function_kwargs_helper(
        config_module, config_space, "data_module_fn", ["data_path"], cast_arguments
    )


def get_model_fn_kwargs(
    config_module, config_space: Dict[str, Any], cast_arguments: Optional[bool] = False
) -> Dict[str, Any]:
    """Returns the kwargs for a ``model_fn`` with defined arguments based on config_space."""
    return _get_function_kwargs_helper(config_module, config_space, "model_fn", [], cast_arguments)


def get_transforms_kwargs(
    config_module, config_space: Dict[str, Any], cast_arguments: Optional[bool] = False
) -> Dict[str, Callable]:
    """Returns the transforms based on config_space."""
    transform_fn_names = [
        "train_transform",
        "train_target_transform",
        "test_transform",
        "test_target_transform",
        "buffer_transform",
        "buffer_target_transform",
    ]
    transforms = {}
    for transform_fn_name in transform_fn_names:
        if hasattr(config_module, transform_fn_name):
            transforms[transform_fn_name] = getattr(config_module, transform_fn_name)(
                **_get_function_kwargs_helper(
                    config_module, config_space, transform_fn_name, [], cast_arguments
                )
            )
    return transforms


def get_metrics_fn_kwargs(
    config_module, config_space: Dict[str, Any], cast_arguments: Optional[bool] = False
) -> Dict[str, Any]:
    """Returns the kwargs for a ``metrics_fn`` with defined arguments based on config_space."""
    return _get_function_kwargs_helper(
        config_module, config_space, "metrics_fn", [], cast_arguments
    )


def _get_function_kwargs_helper(
    config_module,
    config_space: Dict[str, Any],
    function_name: str,
    ignore_args: List[str],
    cast_arguments: Optional[bool] = False,
) -> Dict[str, Any]:
    """Returns kwargs for function based on its interface."""
    all_args = {}
    function_args = get_function_args(config_module, function_name, all_args, ignore_args)
    filtered_args = {key: value for key, value in config_space.items() if key in function_args}
    if cast_arguments:
        filtered_all_args = {key: value for key, value in all_args.items() if key in filtered_args}
        _cast_arguments_to_true_types(arguments=filtered_all_args, args_dict=filtered_args)
    return filtered_args


def get_transforms_dict(
    config_module: ModuleType,
    args: Union[argparse.Namespace, Dict[str, str]],
    function_args: Dict[str, Dict[str, Any]],
) -> Dict[str, Callable]:
    """Creates and returns data transforms for updater."""
    transform_fn_names = [
        "train_transform",
        "train_target_transform",
        "test_transform",
        "test_target_transform",
        "buffer_transform",
        "buffer_target_transform",
    ]
    transforms = {}
    for transform_fn_name in transform_fn_names:
        if hasattr(config_module, transform_fn_name):
            transforms[transform_fn_name] = getattr(config_module, transform_fn_name)(
                **get_function_kwargs(args, function_args[transform_fn_name])
            )
    return transforms


def get_argument_type(arg_spec: inspect.FullArgSpec, argument_name: str) -> Type:
    """Returns the type of the argument with ``argument_name``."""

    def raise_error(arg_type, arg_name):
        raise TypeError(f"Type {arg_type} is not supported (argument {arg_name}).")

    if argument_name not in arg_spec.annotations:
        raise TypeError(f"Missing type annotation for argument {argument_name}.")
    if hasattr(arg_spec.annotations[argument_name], "__origin__"):
        if arg_spec.annotations[argument_name].__origin__ == Union:
            if len(arg_spec.annotations[argument_name].__args__) != 2 or arg_spec.annotations[
                argument_name
            ].__args__[1] != type(None):
                raise_error(arg_spec.annotations[argument_name], argument_name)
            if hasattr(arg_spec.annotations[argument_name].__args__[0], "__origin__"):
                if arg_spec.annotations[argument_name].__args__[0].__origin__ in [list, tuple]:
                    argument_type = arg_spec.annotations[argument_name].__args__[0].__origin__
                else:
                    raise_error(arg_spec.annotations[argument_name], argument_name)
            else:
                argument_type = arg_spec.annotations[argument_name].__args__[0]
            if arg_spec.annotations[argument_name].__origin__ in [list, tuple]:
                argument_type = arg_spec.annotations[argument_name].__origin__
        elif arg_spec.annotations[argument_name].__origin__ in [list, tuple]:
            argument_type = arg_spec.annotations[argument_name].__origin__
        else:
            raise_error(arg_spec.annotations[argument_name], argument_name)
    else:
        argument_type = arg_spec.annotations[argument_name]
    if argument_type not in [int, bool, float, str, list, tuple]:
        raise_error(argument_type, argument_name)
    return argument_type


def to_dense_str(value: Union[bool, List, Tuple, None]) -> str:
    """Converts a variable to string without empty spaces."""
    return str(value).replace(" ", "")


def get_function_args(
    config_module: ModuleType,
    function_name: str,
    all_args: Dict[str, Dict[str, Any]],
    ignore_args: List[str],
) -> List[str]:
    """Returns all argument names of the function and appends new arguments to ``all_args``.

    Args:
        config_module: Renate config file containing functions to access the model and data.
        function_name: Function of which we want to return the arguments.
        all_args: List of all arguments that will be an input to the script.
        ignore_args: Arguments which will not be added to ``all_args`` because these values will
            be created on the fly and are no input to the script.
    """
    if not hasattr(config_module, function_name):
        return []
    arg_spec = inspect.getfullargspec(getattr(config_module, function_name))
    known_args = [arg for arg in arg_spec.args if arg in all_args]
    new_args = [arg for arg in arg_spec.args if arg not in all_args and arg not in ignore_args]
    if arg_spec.defaults is None:
        default_values = {}
    else:
        default_values = {
            arg_spec.args[len(arg_spec.args) - len(arg_spec.defaults) + i]: default
            for i, default in enumerate(arg_spec.defaults)
        }
    for argument_name in known_args:
        argument_type = get_argument_type(arg_spec, argument_name)
        expected_type = all_args[argument_name]["type"]
        if argument_type != expected_type:
            raise TypeError(
                f"Types of `{argument_name}` are not consistent. Defined as type `{argument_type}`"
                f" as well as `{expected_type}`."
            )
        if argument_name not in default_values:
            all_args[argument_name]["required"] = True

    for argument_name in new_args:
        true_type = get_argument_type(arg_spec, argument_name)
        all_args[argument_name] = {
            "type": str if true_type in [bool, list, tuple] else true_type,
            "argument_group": CUSTOM_ARGS_GROUP,
            "true_type": true_type,
        }
        if argument_name in default_values:
            default_value = default_values[argument_name]
            if true_type in [bool, list, tuple] and default_value is not None:
                default_value = to_dense_str(default_value)
            all_args[argument_name]["default"] = default_value
        else:
            all_args[argument_name]["required"] = True
    return arg_spec.args


def get_scheduler_kwargs(
    config_module: ModuleType,
) -> Tuple[Optional[Type[TrialScheduler]], Optional[Dict[str, Any]]]:
    """Creates and returns scheduler type and kwargs for the HPO scheduler."""
    scheduler_fn_name = "scheduler_fn"
    if scheduler_fn_name in vars(config_module):
        return getattr(config_module, scheduler_fn_name)()
    return None, None


parse_by_updater = {
    "ER": _add_experience_replay_arguments,
    "LearningToPrompt": _add_l2p_arguments,
    "LearningToPromptReplay": _add_l2preplay_arguments,
    "SPeft": _add_speft_arguments,
    "DER": _add_dark_experience_replay_arguments,
    "POD-ER": _add_pod_experience_replay_arguments,
    "CLS-ER": _add_cls_experience_replay_arguments,
    "Super-ER": _add_super_experience_replay_arguments,
    "GDumb": _add_gdumb_arguments,
    "Joint": _add_joint_arguments,
    "FineTuning": _add_finetuning_arguments,
    "RD": _add_rd_learner_arguments,
    "Offline-ER": _add_offline_er_arguments,
    "Avalanche-ER": _add_experience_replay_arguments,
    "Avalanche-EWC": _add_avalanche_ewc_learner_arguments,
    "Avalanche-LwF": _add_avalanche_lwf_learner_arguments,
    "Avalanche-iCaRL": _add_avalanche_icarl_learner_arguments,
}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch
from torchvision.models import ResNet18_Weights, resnet18
from torchvision.transforms import GaussianBlur

from renate.benchmark.datasets.vision_datasets import TorchVisionDataModule
from renate.data.datasets import _TransformedDataset
from renate.shift.mmd_detectors import MMDCovariateShiftDetector

# Load CIFAR-10 training dataset.
data_module = TorchVisionDataModule(data_path="data", dataset_name="CIFAR10", val_size=0.2)
data_module.prepare_data()
data_module.setup()
dataset = data_module.train_data()

# We now generate a reference dataset as well as two query datasets: one from the same distribution,
# and one where we simulate a distribution shift by blurring images. In practice, the reference
# dataset should represent your expected data distribution. It could, e.g., be the validation set
# you used during the previous training of your model.
dataset_ref = torch.utils.data.Subset(dataset, list(range(1000)))
dataset_query_in = torch.utils.data.Subset(dataset, list(range(1000, 2000)))
dataset_query_out = torch.utils.data.Subset(dataset, list(range(2000, 3000)))
transform = GaussianBlur(kernel_size=5, sigma=1.0)
dataset_query_out = _TransformedDataset(dataset_query_in, transform)

# Shift detection methods rely on informative (and relatively low-dimensional) features. Here, we
# use a pretrained ResNet model and chop of its output layer. This leads to 512-dimensional
# vectorial features.
feature_extractor = resnet18(weights=ResNet18_Weights.DEFAULT)
feature_extractor.fc = torch.nn.Identity()
feature_extractor.eval()  # Eval mode to use frozen batchnorm stats.

# Now we can instantiate an MMD-based shift detector. We first fit it to our reference datasets,
# and then score both the in-distribution query dataset and the out-of-distribution query dataset.
# In this toy example, the shift is quite obvious and we will see a very high score for the
# out-of-distribution data.
detector = MMDCovariateShiftDetector(feature_extractor=feature_extractor)
print("Fitting detector...")
detector.fit(dataset_ref)
print("Scoring in-distribution data...")
score_in = detector.score(dataset_query_in)
print(f"score = {score_in}")
print("Scoring out-of-distribution data...")
score_out = detector.score(dataset_query_out)
print(f"score = {score_out}")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from typing import Any, Dict, Optional, Union

from transformers import PreTrainedTokenizer

from renate import defaults
from renate.benchmark.datasets.base import DataIncrementalDataModule
from renate.utils.file import download_folder_from_s3
from renate.utils.hf_utils import DataCollatorWithPaddingForWildTime


class WildTimeDataModule(DataIncrementalDataModule):
    """Data module wrapping around the Wild-Time data.

    Huaxiu Yao, Caroline Choi, Bochuan Cao, Yoonho Lee, Pang Wei Koh, Chelsea Finn:
    Wild-Time: A Benchmark of in-the-Wild Distribution Shift over Time. NeurIPS 2022

    Args:
        data_path: the path to the folder containing the dataset files.
        dataset_name: Name of the wild time dataset.
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        time_step: Time slice to be loaded.
        tokenizer: Tokenizer to apply to the dataset. See https://huggingface.co/docs/tokenizers/
            for more information on tokenizers.
        tokenizer_kwargs: Keyword arguments passed when calling the tokenizer's ``__call__``
            function.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_path: Union[Path, str],
        dataset_name: str,
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        time_step: int = 0,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(
            data_path=data_path,
            data_id=time_step,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = dataset_name
        self._tokenizer = tokenizer
        self._tokenizer_kwargs = tokenizer_kwargs

    def prepare_data(self) -> None:
        """Download data.

        If s3 bucket is given, the data is downloaded from s3, otherwise from the original source.
        """
        if self._src_bucket is None:
            from wild_time_data import available_time_steps, load_dataset

            load_dataset(
                dataset_name=self._dataset_name,
                time_step=available_time_steps(self._dataset_name)[0],
                split="train",
                data_dir=self._data_path,
            )
        else:
            from wild_time_data.core import dataset_classes

            dst_dir = Path(self._data_path) / dataset_classes[self._dataset_name].file_name
            if not dst_dir.exists():
                download_folder_from_s3(
                    src_bucket=self._src_bucket,
                    src_object_name=self._src_object_name,
                    dst_dir=str(dst_dir),
                )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        from wild_time_data import available_time_steps, load_dataset

        kwargs = {
            "dataset_name": self._dataset_name,
            "time_step": available_time_steps(self._dataset_name)[self.data_id],
            "data_dir": self._data_path,
            "in_memory": self._dataset_name != "fmow",
            "transform": None if self._dataset_name not in ["fmow", "yearbook"] else lambda x: x,
        }
        if self._tokenizer:
            kwargs["transform"] = lambda x: self._tokenizer(x, **(self._tokenizer_kwargs or {}))
        train_data = load_dataset(split="train", **kwargs)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = load_dataset(split="test", **kwargs)
        if self._dataset_name in ["huffpost", "arxiv"]:
            self._train_collate_fn = DataCollatorWithPaddingForWildTime(tokenizer=self._tokenizer)
            self._val_collate_fn = DataCollatorWithPaddingForWildTime(tokenizer=self._tokenizer)
            self._test_collate_fn = DataCollatorWithPaddingForWildTime(tokenizer=self._tokenizer)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.shift.kernels import RBFKernel


@pytest.mark.parametrize("kernel", [RBFKernel()])
@pytest.mark.parametrize(
    "X1, X2",
    [
        (torch.randn(20, 2), torch.randn(20, 2)),
        (torch.randn(10, 2), torch.randn(20, 2)),
        (torch.randn(20, 2), torch.randn(10, 2)),
    ],
)
def test_kernel_shapes(kernel, X1, X2):
    K = kernel(X1, X2)
    assert K.size() == (X1.size(0), X2.size(0))


@pytest.mark.parametrize("kernel", [RBFKernel()])
def test_kernel_shape_mismatch(kernel):
    with pytest.raises(Exception):
        kernel(torch.randn(10, 2), torch.randn(20, 3))


def test_rbf_vs_manual_computation():
    X0 = torch.tensor([[0.0, 0.0], [0.0, 1.0]])
    X1 = torch.tensor([[1.0, 0.0], [2.0, 2.0]])
    kernel = RBFKernel(lengthscale=1.0)
    K = kernel(X0, X1)
    K_exp = torch.exp(-0.5 * torch.tensor([[1.0, 8.0], [2.0, 5.0]]))
    assert torch.allclose(K, K_exp)


def test_rbf_kernel_limits():
    """Tests limit behavior of RBF kernel"""
    X = torch.randn(10, 2)
    # Small lengthscales should result in vanishing off-diagonal terms.
    kernel = RBFKernel(lengthscale=1e-8)
    K = kernel(X, X)
    assert torch.allclose(K, torch.eye(10))
    # High lengthscales should result in a matrix of all ones.
    kernel = RBFKernel(lengthscale=1e8)
    K = kernel(X, X)
    assert torch.allclose(K, torch.ones(10))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy

import pytest
import torch

import renate.defaults as defaults
from renate.updaters.experimental.fine_tuning import FineTuningModelUpdater


def get_model_and_dataset():
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    dataset = torch.utils.data.TensorDataset(
        torch.rand((100, 10)),
        torch.randint(10, (100,)),
    )
    return model, dataset


@pytest.mark.parametrize("devices", [None, 1])
@pytest.mark.parametrize("strategy", [None, "ddp"])
@pytest.mark.parametrize("accelerator", ["cpu"])
def test_fine_tuning_updater(devices, strategy, accelerator):
    """This test checks that the memory buffer is updated correctly."""
    init_model, dataset = get_model_and_dataset()

    model = copy.deepcopy(init_model)

    model_updater = FineTuningModelUpdater(
        model,
        optimizer=pytest.helpers.get_partial_optimizer(),
        loss_fn=pytest.helpers.get_loss_fn(),
        max_epochs=1,
        devices=devices,
        strategy=strategy,
        accelerator=accelerator,
    )
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)

    for p1, p2 in zip(init_model.parameters(), model.parameters()):
        assert p1.data.ne(p2.data).sum().item() != 0



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from typing import Any

import torch
from torch import Tensor


class PredictionStrategy(ABC):
    @abstractmethod
    def __call__(self, inputs: Tensor, training: bool, **kwargs: Any) -> Tensor:
        return inputs


class ICaRLClassificationStrategy(PredictionStrategy):
    def __call__(self, inputs: Tensor, training: bool, class_means: Tensor) -> Tensor:
        if training:
            return super().__call__(inputs, training)
        normalized_inputs = (inputs.T / torch.norm(inputs.T, dim=0)).T
        return (-torch.cdist(class_means.to(normalized_inputs.device)[:, :].T, normalized_inputs)).T



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pytest

import renate.defaults as defaults
from renate.updaters.experimental.joint import JointLearner


def test_joint_learner_memory_append():
    """This test checks that the memory buffer is updated correctly."""
    model, dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        num_hidden_layers=3,
        hidden_size=32,
        train_num_samples=100,
        test_num_samples=100,
    )
    dataset_len = len(dataset)
    model_updater = pytest.helpers.get_simple_updater(
        model=model,
        partial_optimizer=pytest.helpers.get_partial_optimizer(),
        learner_class=JointLearner,
        learner_kwargs={},
        max_epochs=1,
    )
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    assert len(model_updater._learner._memory_buffer) == dataset_len
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    assert len(model_updater._learner._memory_buffer) == 2 * dataset_len



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import argparse
import logging
import shutil
import sys
from pathlib import Path

from pytorch_lightning import seed_everything

from renate import defaults
from renate.cli.parsing_functions import (
    get_function_kwargs,
    get_transforms_dict,
    get_updater_and_learner_kwargs,
    parse_arguments,
)
from renate.utils.file import maybe_download_from_s3, move_to_uri
from renate.utils.module import (
    get_and_setup_data_module,
    get_learning_rate_scheduler,
    get_loss_fn,
    get_metrics,
    get_model,
    get_optimizer,
    import_module,
)
from renate.utils.optimizer import create_partial_optimizer
from renate.utils.syne_tune import redirect_to_tmp

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ModelUpdaterCLI:
    """Entry point to update a model.

    Given a dataset, a model and a training configuration, this class will update the model with the
    given data.

    It will create following folder structure:
        args.working_dir
            data (filled by the DataModule)
            input_state_dir (content provided in args.state_url is copied to this location)
                learner.pkl
                model.pt
            output_state_dir (outcome of the model update)
                learner.pkl
                model.pt
    """

    def _copy_state_to_working_directory(
        self, input_state_url: str, input_state_folder: str
    ) -> None:
        """Copies current state into the working directory.

        If state is on s3, download directly from there to `input_state_folder`.
        If state is already in local directory, copy to right folder unless it is already in the
        right folder. If `input_state_folder` exists but `input_state_url` is not provided,
        `input_state_folder` will be removed.
        """
        local_dir = maybe_download_from_s3(input_state_url, input_state_folder)
        folder_downloaded_from_s3 = local_dir == input_state_folder
        if not folder_downloaded_from_s3:
            if local_dir != input_state_folder:
                shutil.rmtree(input_state_folder, ignore_errors=True)
                if local_dir is not None:
                    shutil.copytree(
                        local_dir,
                        input_state_folder,
                        ignore=shutil.ignore_patterns("*.sagemaker-uploading"),
                        dirs_exist_ok=True,
                    )

    def _prepare_data_state_model(self, args: argparse.Namespace) -> None:
        """Assigns locations for data, state and model."""
        self._data_folder = defaults.data_folder(args.working_directory)
        self._input_state_folder = defaults.input_state_folder(args.working_directory)
        working_directory = redirect_to_tmp(args.st_checkpoint_dir or args.working_directory)
        self._output_state_folder = defaults.output_state_folder(working_directory)
        self._current_model_file = defaults.model_file(self._input_state_folder)
        self._next_model_file = defaults.model_file(self._output_state_folder)
        self._copy_state_to_working_directory(args.input_state_url, self._input_state_folder)
        if not Path(self._input_state_folder).is_dir():
            self._input_state_folder = None
        if not Path(self._current_model_file).is_file():
            self._current_model_file = None

    def run(self):
        config_module = None
        for i, arg_value in enumerate(sys.argv):
            if arg_value == "--config_file":
                config_module = import_module("config_module", sys.argv[i + 1])
                break
        if config_module is None:
            raise RuntimeError("The following argument is required: --config_file")
        args, function_args = parse_arguments(
            config_module=config_module,
            function_names=[
                "model_fn",
                "data_module_fn",
                "train_transform",
                "test_transform",
                "buffer_transform",
                "scheduler_fn",
                "loss_fn",
                "optimizer_fn",
                "lr_scheduler_fn",
                "metrics_fn",
            ],
            ignore_args=["data_path", "model_state_url"],
        )

        seed_everything(args.seed, True)
        self._prepare_data_state_model(args)

        data_module = get_and_setup_data_module(
            config_module,
            data_path=self._data_folder,
            prepare_data=args.prepare_data,
            **get_function_kwargs(args=args, function_args=function_args["data_module_fn"]),
        )
        model = get_model(
            config_module,
            model_state_url=None if getattr(args, "reset", False) else self._current_model_file,
            **get_function_kwargs(args=args, function_args=function_args["model_fn"]),
        )
        loss_fn = get_loss_fn(
            config_module,
            not args.updater.startswith("Avalanche-"),
            **get_function_kwargs(args=args, function_args=function_args["loss_fn"]),
        )
        partial_optimizer = get_optimizer(
            config_module,
            **get_function_kwargs(args=args, function_args=function_args["optimizer_fn"]),
        )
        if partial_optimizer is None:
            partial_optimizer = create_partial_optimizer(
                optimizer=args.optimizer,
                lr=args.learning_rate,
                momentum=args.momentum,
                weight_decay=args.weight_decay,
            )
        lr_scheduler_config = get_learning_rate_scheduler(
            config_module,
            **get_function_kwargs(args=args, function_args=function_args["lr_scheduler_fn"]),
        )
        lr_scheduler_kwargs = {}
        if lr_scheduler_config is not None:
            lr_scheduler_kwargs["learning_rate_scheduler"] = lr_scheduler_config[0]
            lr_scheduler_kwargs["learning_rate_scheduler_interval"] = lr_scheduler_config[1]
        metrics = get_metrics(
            config_module,
            **get_function_kwargs(args=args, function_args=function_args["metrics_fn"]),
        )

        model_updater_class, learner_kwargs = get_updater_and_learner_kwargs(args)

        model_updater = model_updater_class(
            model=model,
            optimizer=partial_optimizer,
            input_state_folder=self._input_state_folder,
            output_state_folder=self._output_state_folder,
            max_epochs=args.max_epochs,
            metric=args.metric,
            mode=args.mode,
            logged_metrics=metrics,
            accelerator=args.accelerator,
            devices=args.devices,
            precision=args.precision,
            strategy=args.strategy,
            gradient_clip_algorithm=args.gradient_clip_algorithm,
            gradient_clip_val=args.gradient_clip_val,
            early_stopping_enabled=args.early_stopping,
            deterministic_trainer=args.deterministic_trainer,
            loss_fn=loss_fn,
            **learner_kwargs,
            **lr_scheduler_kwargs,
            **get_transforms_dict(config_module, args, function_args),
        )

        model_updater.update(
            train_dataset=data_module.train_data(),
            val_dataset=data_module.val_data(),
            train_dataset_collate_fn=data_module.train_collate_fn(),
            val_dataset_collate_fn=data_module.val_collate_fn(),
            task_id=args.task_id,
        )

        if args.output_state_url is not None:
            move_to_uri(self._output_state_folder, args.output_state_url)


if __name__ == "__main__":
    ModelUpdaterCLI().run()
